from os import path
from pyppl import Box
from collections import OrderedDict
from bioprocs.utils import shell2 as shell, logger
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

snpfile   = {{ i.snpfile | quote}}
outfile   = {{ o.outfile | quote}}
inopts    = {{ args.inopts | repr}}
snpcol    = {{ args.snpcol or 0 | repr}}
dbsnp     = {{ args.dbsnp | quote}}
vcftools  = {{ args.vcftools | quote}}
sortby    = {{ args.sortby | quote}}
jobindir  = {{ job.indir | quote}}
joboutdir = {{ job.outdir | quote}}
ncol      = {{ args.ncol | repr}}

# cruzdb
tool      = {{ args.tool | quote}}
genome    = {{ args.genome | quote}}
dbsnpver  = {{ args.dbsnpver | quote}}

if tool == 'cruzdb':
	'''
	snp151(
		bin=1566,
		chrom='chr8',
		chromStart=128700232L,
		chromEnd=128700233L,
		name='rs7005394',
		score=0,
		strand=u'+',
		refNCBI='T',
		refUCSC='T',
		observed='C/T',
		molType=u'genomic',
		class=u'single',
		valid=set(['by-2hit-2allele', 'by-cluster', 'by-hapmap', 'by-frequency', 'by-1000genomes']),
		avHet=0.498652,
		avHetSE=0.025923,
		func=set(['ncRNA']),
		locType=u'exact',
		weight=1L,
		exceptions=set([]),
		submitterCount=26,
		submitters='1000GENOMES,ABI,BCM-HGSC-SUB,BCM_SSAHASNP,BGI,BL,BUSHMAN,COMPLETE_GENOMICS,DDI,ENSEMBL,EVA-GONL,EVA_DECODE,EVA_GENOME_DK,EVA_UK10K_ALSPAC,EVA_UK10K_TWINSUK,GMI,HAMMER_LAB,HGSV,HUMANGENOME_JCVI,ILLUMINA-UK,JMKIDD_LAB,PJP,SSAHASNP,SSMP,TISHKOFF,WEILL_CORNELL_DGM,',
		alleleFreqCount=2,
		alleles='C,T,',
		alleleNs='2634.000000,2374.000000,',
		alleleFreqs='0.525958,0.474042,',
		bitfields=set(['maf-5-all-pops', 'maf-5-some-pop'])
	)
	'''

	# snps
	reader  = TsvReader(snpfile, **inopts)
	snplist = list(set(r[snpcol] for r in reader))
	reader.close()
	nsnps   = len(snplist)
	logger.info('Got %s snps.', nsnps)

	from cruzdb import Genome
	g          = Genome(genome)
	outfiletmp = outfile + '.tmp'
	writer     = TsvWriter(outfiletmp)
	chunksize  = 1000
	logger.info('Doing with chunks with %s in each one', chunksize)
	for i in range(0, nsnps, chunksize):
		logger.info('- Handling batch with index starts with %s (%.2f%%) ...', i, 100.0*float(i)/float(nsnps))
		chunk = snplist[i:i+chunksize]
		sql   = 'SELECT chrom, chromStart, chromEnd, name, score, strand, refUCSC, alleles, alleleFreqs FROM snp{dbsnpver} WHERE name in ({snps})'.format(
			dbsnpver = dbsnpver,
			snps     = ', '.join("'{}'".format(s) for s in chunk)
		)
		result = g.sql(sql)
		for r in result:
			if ncol == 6:
				writer.write([
					r.chrom, r.chromStart, r.chromEnd, r.name, r.score, r.strand
				])
			else:
				allfreqs = OrderedDict(sorted(zip(r.alleles.rstrip(',').split(','), r.alleleFreqs.rstrip(',').split(',')), key = lambda x: float(x[1]), reverse = True))
				reffreq  = allfreqs.get(r.refUCSC, '0')
				if r.refUCSC in allfreqs:
					del allfreqs[r.refUCSC]
				if '' in allfreqs:
					del allfreqs['']

				if ncol == 9:
					writer.write([
						r.chrom, r.chromStart, r.chromEnd, r.name, r.score, r.strand,
						r.refUCSC, ','.join(allfreqs.keys()), ','.join([reffreq] + list(allfreqs.values()))
					])
				else: #8
					writer.write([
						r.chrom, r.chromStart, r.chromEnd, r.name, r.score, r.strand,
						r.refUCSC, ','.join(allfreqs.keys())
					])
	writer.close()

else:
	# snps
	snplist = path.join(jobindir, path.basename(snpfile) + '.list')
	reader  = TsvReader(snpfile, **inopts)
	writer  = TsvWriter(snplist)
	for r in reader:
		writer.write([r[snpcol]])
	reader.close()
	writer.close()

	shell.load_config(vcftools = vcftools)

	params = Box()
	params.snps = snplist
	params.recode = True
	params.out = path.join(joboutdir, 'tmp')
	if dbsnp.endswith('.gz'):
		params.gzvcf = dbsnp
	elif not path.isfile(dbsnp):
		raise ValueError('dbsnp file (args.dbsnp) is required by tool "local"')
	else:
		params.vcf = dbsnp
	shell.fg.vcftools(**params)

	reader = TsvReader(params.out + '.recode.vcf', cnames = False)
	outfiletmp = outfile + '.tmp'
	writer = TsvWriter(outfiletmp)
	for r in reader:
		if ncol == 6:
			writer.write([r[0], r[1], r[1], r[2], 0, '+'])
		elif ncol == 8:
			writer.write([r[0], r[1], r[1], r[2], 0, '+', r[3], r[4]])
		else:
			ValueError('Available ncol is 6 or 8 for `args.tool = "dbsnp"`')
	reader.close()
	writer.close()

if sortby == 'coord':
	shell.sort(k = ['1,1', '2,2n'], _ = outfiletmp, _stdout = outfile)
elif sortby == 'name':
	shell.sort(k = '4', _ = outfiletmp, _stdout = outfile)
else:
	shell.mv(outfiletmp, outfile)

if tool == 'cruzdb':
	shell.rm_rf(outfiletmp)
else:
	shell.rm_rf(outfiletmp, params.out + '.recode.vcf')



from os import path
from pyppl import Box
from bioprocs.utils import ensureBox, shell2 as shell
from bioprocs.utils.reference import vcfIndex

infile = {{i.infile | quote}}
outdir = {{o.outdir | quote}}
plink  = {{args.plink | quote}}
tabix  = {{args.tabix | quote}}
params = {{args.params}}
params = ensureBox(params)

# resolve plink 1.x --set-missing-var-ids doesn't distinguish $1, $2,... for ref and alts
if 'set-missing-var-ids' in params and "$" in params['set-missing-var-ids']:
	tmpfile = path.join(outdir, 'withvarids.vcf')
	import vcf
	reader = vcf.Reader(filename = infile)
	stream = open(tmpfile, 'w')
	writer = vcf.Writer(stream, reader)
	for r in reader:
		if not r.ID:
			all_alts = [r.REF] + list(r.ALT)
			r.ID     = params['set-missing-var-ids'].replace(
				'@', r.CHROM).replace('#', str(r.POS))
			for i, a in enumerate(all_alts):
				r.ID = r.ID.replace('${}'.format(i+1), str(a))
		writer.write_record(r)
	stream.close()
	infile = tmpfile
infile = vcfIndex(infile, tabix = tabix)

params.vcf = infile
params['make-bed'] = True
params.out = path.join(outdir, {{i.infile | fn2 | quote}})

shell.fg.plink(**params)



import threading
from time import sleep
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from bioprocs.utils.shell2 import runcmd
from traceback import format_exc
from queue import Queue

def distribute(total, nthread):
	"""
	Try to distribute jobs into N threads as equal as possible.
	For example: distributing 10 jobs on 3 threads, we prefer (4, 3, 3) than (4, 4, 2)
	How to do it?
	1. get the ceiling size for each thread, that should be (3, 3, 3), from `divmod(10, 3)[0]`
	2. get the modulo by `divmod(10, 3)[1]`, which is the first # nthreads to add one more job to
	@params:
		`total`: The total # jobs
		`nthread`: The # threads
	@returns:
		A list of # jobs distribute on each thread.
	"""
	nthread = min(total, nthread)
	(m, d) = divmod(total, nthread)
	ret = [m] * nthread
	for i in range(d):
		ret[i] += 1
	return ret

def distributeList(joblist, nthread):
	lists = distribute(len(joblist), nthread)
	start = 0
	for l in lists:
		yield joblist[start:l+start]
		start += l

class Parallel(object):

	def __init__(self, nthread = 1, backend = 'thread', raiseExc = False):
		PoolExecutor   = ProcessPoolExecutor if backend.lower() in 'multiprocessing' else ThreadPoolExecutor
		self.executor  = PoolExecutor(max_workers = nthread)
		self.raiseExc  = raiseExc

	def __del__(self):
		if self.executor:
			self.executor.shutdown()

	@staticmethod
	def _run(func, arg):
		if callable(func):
			ret = func(*arg)
			return ret
		else:
			return runcmd(func.format(*arg))

	def run(self, func, args):
		submits   = []
		results   = []
		exception = None
		excno     = 0
		for arg in args:
			submits.append(self.executor.submit(Parallel._run, func, arg))

		for submit in submits:
			try:
				results.append(submit.result())
			except Exception as ex:
				#results.append(None)
				exception = type(ex), format_exc()
				excno    += 1

		if excno > 0 and self.raiseExc:
			raise exception[0](exception[1])

		return results

class LargeFileHandler(object):
	HAND = -1

	def __init__(self, reader, compute = None, summarize = None, compute_multi = None, summarize_multi = None, size = 1000, readfunc = lambda r: next(r), nthread = 1, raiseExc = True):
		LargeFileHandler.HAND = -1
		self.reader           = reader
		self.readfunc         = readfunc
		self.raiseExc         = raiseExc
		self.size             = size
		self.queue            = Queue()
		self.nthread          = nthread
		self.compute          = compute
		self.summarize        = summarize
		self.compute_multi    = compute_multi
		self.summarize_multi  = summarize_multi
		self.index            = 0
		self.lock             = threading.Lock()
		self.stop             = False
		if self.compute and self.compute_multi:
			raise ValueError('Only one of compute and compute_multi is needed.')
		if self.summarize and self.summarize_multi:
			raise ValueError('Only one of summarize and summarize_multi is needed.')
		if not callable(self.compute) and not callable(self.compute_multi):
			raise ValueError('One of compute and compute_multi is needed.')
		if not callable(self.summarize) and not callable(self.summarize_multi):
			raise ValueError('One of summarize and summarize_multi is needed.')

		for _ in range(nthread):
			self.producer()

	def producer(self):
		if self.stop: return
		with self.lock:
			lines = []
			for _ in range(self.size):
				try:
					line = self.readfunc(self.reader)
					if line is False: break
					if line is None:
						self.stop = True
						break
					lines.append(line)
				except StopIteration:
					break
			if lines:
				self.queue.put((self.index, lines))
				self.index += 1

	def run(self):
		for _ in range(self.nthread):
			t = threading.Thread(target = self.worker)
			t.daemon = True
			t.start()
		self.queue.join()

	def worker(self):
		while not self.queue.empty():
			index, data = self.queue.get()
			try:
				if self.compute:
					computed = [self.compute(d) for d in data]
				else:
					computed = self.compute_multi(data)
				while LargeFileHandler.HAND + 1 != index:
					sleep(.01)
				if self.summarize:
					for c in computed:
						self.summarize(c)
				else:
					self.summarize_multi(computed)
				self.producer()
			except:
				if self.raiseExc:
					raise
			finally:
				LargeFileHandler.HAND = index
			self.queue.task_done()



from pyppl import Box
from bioprocs.utils import shell
from bioprocs.utils.shell import Shell

infile   = {{ i.infile | quote}}
outfile  = {{ o.outfile | quote }}
outidx   = {{ o.outidx | quote }}
tool     = {{ args.tool | quote}}
samtools = {{ args.samtools | quote}}
params   = {{ args.params | repr}}
nthread  = {{ args.nthread | repr}}

shell.TOOLS['samtools'] = samtools

shell.ln_s(infile, outfile)

def run_samtools():
	shellst = Shell(subcmd = True).samtools
	params['@'] = nthread
	shellst.index(_ = [infile, outidx]).run()

tools = dict(
	samtools = run_samtools
)

try:
	tools[tool]()
except KeyError:
	raise KeyError('Tool {!r} is not supported.'.format(tool))
except:
	raise
finally:
	pass



from shutil import move, rmtree
from os import path, symlink, remove
from sys import stderr
from pyppl import Box
from bioprocs.utils import mem2, shell2 as shell

infile     = {{ i.infile | quote }}
inprefix   = {{ i.infile | fn | quote}}
outfile    = {{ o.outfile | quote }}
joboutdir  = {{job.outdir | quote}}
infmt      = {{ i.infile | ext | [1:] | quote}}
tmpdir     = {{ args.tmpdir | quote}}
tmpdir     = path.join(tmpdir, {{proc.id, fn(i.infile), str(job.index+1) | @join: "_" | quote }})
steps      = {{ args.steps | repr }}
params     = {{ args.params | repr }}
argsmem    = {{ args.mem | repr }}
sortby     = {{ args.sortby | quote}}
nthread    = {{ args.nthread | repr}}
ref        = {{ args.ref | repr}}
knownSites = {{ args.knownSites | repr}}
biobambam  = {{ args.biobambam | quote }}
elprep     = {{ args.elprep | quote}}
sambamba   = {{ args.sambamba | quote}}
samtools   = {{ args.samtools | quote}}
picard     = {{ args.picard | quote}}
tool       = {{ args.tool | quote }}


shell.load_config(
	biobambam = biobambam,
	sambamba  = sambamba,
	samtools  = samtools,
	picard    = picard,
	elprep    = elprep
)

if steps.rmdup:
	steps.markdup = True
if not path.exists(tmpdir):
	shell.mkdir(tmpdir, p = True)
if steps.recal and tool != 'elprep':
	raise ValueError('Step "recal" is only enabled by "elprep", use pBamRecal for other tools.')


def run_biobambam():
	mem = mem2(argsmem, 'M')
	if steps.index:
		params.index         = 1
		params.indexfilename = outfile + '.bai'

	params.I              = infile
	params.O              = outfile
	params.SO             = sortby
	params.blockme        = mem
	params.tmpfile        = path.join(tmpdir, 'biobambam.tmp')
	params.inputformat    = infmt
	params.outfmt         = 'bam'
	params.inputthreads   = nthread
	params.outputthreads  = nthread
	params.markduplicates = int(steps.markdup)
	params.rmdup          = int(steps.rmdup)

	shell.fg.biobambam(**params)

def run_sambamba():
	global infile
	if not (steps.sort or steps.index or steps.markdup or steps.rmdup):
		shell.fg.sambamba.view(S = True, f = 'bam', o = outfile, t = nthread, _ = infile)
	else:
		bamfile = outfile
		if infmt == 'sam':
			bamfile = path.join(joboutdir, inprefix + '.s2b.bam')
			shell.fg.sambamba.view(S = True, f = 'bam', o = bamfile, t = nthread, _ = infile)
			infile = bamfile
		if steps.sort:
			if sortby == 'queryname':
				params.n = True
				params.N = True
			bamfile       = path.join(joboutdir, inprefix + '.sorted.bam')
			params.m      = argsmem
			params.tmpdir = tmpdir
			params.o      = bamfile
			params.t      = nthread
			params._      = infile
			shell.fg.sambamba.sort(**params)
			if infile != {{i.infile | quote}}:
				shell.rm_rf(infile)
			infile = bamfile
		if steps.markdup:
			bamfile = path.join(joboutdir, inprefix + '.dedup.bam')
			shell.fg.sambamba.markdup(r = steps.rmdup, t = nthread, tmpdir = tmpdir,
				_ = [infile, bamfile])
			if infile != {{i.infile | quote}}:
				shell.rm_rf(infile)
			infile = bamfile
		if steps.index:
			if path.exists(infile + '.bai'):
				shell.mv(infile + '.bai', outfile + '.bai')
			else:
				shell.fg.sambamba.index(t = nthread, _ = [infile, infile + '.bai'])
		if infile != outfile:
			if path.exists(infile + '.bai'):
				shell.mv(infile + '.bai', outfile + '.bai')
			shell.mv(infile, outfile)

def run_samtools():
	global infile
	if not (steps.sort or steps.index or steps.markdup or steps.rmdup):
		shell.fg.samtools.view(b = True, o = outfile, O = 'bam', _ = infile)
	else:
		bamfile = outfile
		if steps.sort:
			mem = mem2(argsmem, 'M')
			bamfile = path.join(joboutdir, inprefix + '.sorted.bam')
			shell.fg.samtools.sort(
				m     = mem + 'M',
				n     = sortby == 'queryname',
				o     = bamfile,
				T     = tmpdir,
				O     = 'bam',
				_     = infile,
				**{'@': nthread}
			)
			if infile != {{i.infile | quote}}:
				shell.rm_rf(infile)
			infile = bamfile
		if steps.markdup or steps.rmdup:
			bamfile = path.join(joboutdir, inprefix + '.dedup.bam')
			shell.fg.samtools.rmdup(infile, bamfile)
			if infile != {{i.infile | quote}}:
				shell.rm_rf(infile)
			infile = bamfile
		if steps.index:
			shell.samtools.index(bamfile, outfile + '.bai')
		if infile != outfile:
			if path.exists(infile + '.bai'):
				shell.mv(infile + '.bai', outfile + '.bai')
			shell.mv(infile, outfile)

def run_picard():
	global infile
	mem = mem2(argsmem, '-jdict')
	mem['-Djava.io.tmpdir'] = tmpdir
	if not (steps.sort or steps.index or steps.markdup or steps.rmdup):
		shell.fg.picard.SamFormatConverter(TMP_DIR = tmpdir, I = infile, O = outfile)
	else:
		bamfile = outfile
		if steps.sort:
			bamfile = path.join(joboutdir, inprefix + '.sorted.bam')
			shell.fg.picard.ShortSam(TMP_DIR = tmpdir, I = infile, O = bamfile, SO = sortby)
			if infile != {{i.infile | quote}}:
				shell.rm_rf(infile)
			infile = bamfile
		if steps.markdup:
			mfile = "/dev/null"
			bamfile = path.join(joboutdir, inprefix + '.dedup.bam')
			shell.fg.picard.MarkDuplicates(REMOVE_DUPLICATES = 'true' if steps.rmdup else 'false',
				TMP_DIR = tmpdir, I = infile, O = bamfile, M = mfile)
			if infile != {{i.infile | quote}}:
				shell.rm_rf(infile)
			infile = bamfile
		if steps.index:
			shell.fg.picard.BuildBamIndex(TMP_DIR = tmpdir, I = infile, O = outfile + '.bai')
		if infile != outfile:
			if path.exists(infile + '.bai'):
				shell.mv(infile + '.bai', outfile + '.bai')
			shell.mv(infile, outfile)

def run_elprep():

	params['log-path']                       = joboutdir
	params['nr-of-threads']                  = nthread
	params['sorting-order']                  = sortby if steps.sort else 'keep'
	params['mark-duplicates']                = steps.markdup
	params['remove-duplicates']              = steps.rmdup
	params['intermediate-files-output-type'] = 'bam'
	params['contig-group-size']              = 1
	params['tmp-path']                       = tmpdir
	params['']                               = [infile, outfile]
	if steps.markdup:
		params['mark-optical-duplicates'] = path.join(joboutdir, inprefix + '.opticaldups.txt')
	if steps.recal:
		params['bqsr']           = path.join(joboutdir, inprefix + '.bqsr.txt')
		params['bqsr-reference'] = ref + '.elprep'
		if knownSites:
			params['known-sites'] = knownSites

	shell.fg.elprep.sfm(**params)
	if steps.index:
		shell.samtools.index(outfile, outfile + '.bai')

tools = {
	'biobambam': run_biobambam,
	'sambamba' : run_sambamba,
	'samtools' : run_samtools,
	'picard'   : run_picard,
	'elprep'   : run_elprep
}

try:
	tools[tool]()
except KeyError:
	raise KeyError('Tool {!r} not supported.'.format(tool))
except Exception as ex:
	stderr.write ("Job failed: %s" % str(ex))
	raise
finally:
	rmtree (tmpdir)



import vcf
from bioprocs.utils.tsvio import TsvWriter, TsvRecord

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}

reader = vcf.Reader(filename = infile)
writer = TsvWriter(outfile)
writer.meta.add(
	'mutation_id',
	'ref_counts',
	'var_counts',
	'normal_cn',
	'minor_cn',
	'major_cn',
	'variant_case',
	'variant_freq',
	'genotype'
)
while True:
	try:
		r = next(reader)
		record = TsvRecord()
		record.normal_cn = 2
		# mutation_id	ref_counts	var_counts	normal_cn	minor_cn	major_cn	variant_case	variant_freq	genotype
		# NA12156:BB:chr2:175263063	3812	14	2	0	2	NA12156	0.0036591741	BB
		for samdata in record.samples:
			

	except StopIteration:
		break
	except:
		continue



"""
usage: All-FIT.py [-h] -i INPUTDIRFILE -d OUTPUTDIR -o OUTPUTNAME
				  [-s STANDARDDEVIATION] [-t {somatic,all}]

This is a program that estimates specimen purity from tumor-only sample
sequenced with deep sequencing, called All-FIT (Allele Frequency based
Imputation of Tumor Purity). We do not provide model for a variant that is
being called on chrX (male) or chrY. Users may consider the option of removing
all variants on chrX (male) and on chrY, if there is any in the input sample.
Input sample should be a tab-delimited file with 4 columns and a header of ID
Allele_Freq Depth Ploidy

ID	Allele_Freq	Depth	Ploidy
Var1	39.0	378	2
Var2	19.0	996	2
Var3	64.0	849	1
Var4	77.0	487	2
Var5	49.0	639	2
Var6	52.0	891	2
Var7	52.0	841	2

optional arguments:
  -h, --help            show this help message and exit
  -i INPUTDIRFILE, --inputDirFile INPUTDIRFILE
						Input file name with path
  -d OUTPUTDIR, --outputDir OUTPUTDIR
						Output directory
  -o OUTPUTNAME, --outputName OUTPUTNAME
						Output file name prefix
  -s STANDARDDEVIATION, --standardDeviation STANDARDDEVIATION
						How many standard deviation or confidence interval of
						estimated purity
  -t {somatic,all}, --typeOfInput {somatic,all}
						Types of variants input whether germline variants are
						removed(somatic) or not(all)
"""
from os import path
from pyppl import Box
from bioprocs.utils import logger, shell2 as shell
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter, TsvRecord

infile   = {{i.infile | quote}}
cnfile   = {{i.cnfile | quote}}
outstem  = {{o.outfile | stem | quote}}
outdir   = {{o.outdir | quote}}
allfit   = {{args.allfit | quote}}
bcftools = {{args.bcftools | quote}}
bedtools = {{args.bedtools | quote}}
params   = {{args.params | repr}}
mutctrl  = {{args.mutctrl | repr}}
cnctrl   = {{args.cnctrl | repr}}
nthread  = {{args.nthread | repr}}

shell.load_config(allfit = dict(
	_exe = allfit,
	_env = dict(
		OPENBLAS_NUM_THREADS = str(nthread),
		OMP_NUM_THREADS      = str(nthread),
		NUMEXPR_NUM_THREADS  = str(nthread),
		MKL_NUM_THREADS      = str(nthread)
	)),
	bcftools = bcftools, bedtools = bedtools)

def MutVCF2BED(vcffile):
	"""
	Convert single(paired)-sample VCFs to a BED file like:
	#CHROM  START   END	    NAME      AF  Depth GT
	chr1    70820   70820   chr:70820 .39 1083  1
	"""
	ret = None
	samples = shell.bcftools.query(l = vcffile).strip().splitlines()
	assert len(samples) <= 2
	samidx  = 0 if len(samples) == 1 or mutctrl in (1, -1, None) else 1
	sample  = samples[samidx]
	ret = path.join(outdir, sample + '.muts.bed')
	writer = TsvWriter(ret)
	writer.cnames = ['CHROM', 'START', 'END', 'NAME', 'AF', 'Depth', 'CN']
	for line in shell.bcftools.query(
		_iter = True,
		_ = vcffile,
		s = sample,
		f = '%CHROM\t%POS\t%POS\t%CHROM:%POS\t[%AF]\t[%AD{1}]\t[%GT]\n'):
		items = line.strip().split('\t')
		try:
			gt = int(items[-1][0]) + int(items[-1][-1])
		except (TypeError, ValueError):
			logger.warning('Variant %s dropped failure fetching genotype' % (items[3]))
			continue
		items[-1] = gt
		writer.write(items)
	writer.close()
	return ret

def CnVCF2BED(vcffile):
	"""
	Convert single(paired)-sample VCFs to a BED file like:
	#CHROM  START   END	    CN
	chr1    70820   70820   5
	"""

	ret = {}
	samples = shell.bcftools.query(l = vcffile).strip().splitlines()
	assert len(samples) <= 2
	samidx  = 0 if len(samples) == 1 or cnctrl in (1, -1, None) else 1
	sample  = samples[samidx]
	ret = path.join(outdir, sample + '.cn.bed')
	writer = TsvWriter(ret)
	writer.cnames = ['CHROM', 'START', 'END', 'CN']
	for line in shell.bcftools.query(
		_iter = True,
		_ = vcffile,
		s = sample,
		f = '%CHROM\t%POS\t[%END]\t[%CN]\n'):
		items = line.strip().split('\t')
		if items[3] == '0':
			logger.warning('Record skipped with CN=0: %s', line)
			continue
		writer.write(items)
	writer.close()
	return ret

if infile.endswith('.vcf') or infile.endswith('.vcf.gz'):
	mutbed = MutVCF2BED(infile)
	if cnfile.endswith('.vcf') or cnfile.endswith('.vcf.gz'):
		cnbed = CnVCF2BED(cnfile)
	elif cnfile.endswith('.bed'):
		cnbed = cnfile
	else:
		raise ValueError('Unsupported CNV file.')

	infile = path.join(outdir, path.splitext(mutbed)[0] + '.allfit.txt')
	writer = TsvWriter(infile)
	writer.cnames = ['ID', 'Allele_Freq', 'Depth', 'Ploidy']
	writer.writeHead()
	for line in shell.bedtools.intersect(
		_iter = True,
		a = mutbed,
		b = cnbed,
		loj = True):
		# CHROM  START END NAME AF   Depth       GT
		# 0      1     2   3    4        5        6
		# CHROM START END CN
		# 7     8     9   10
		parts = line.strip().split('\t')
		writer.write([
			parts[3],
			float(parts[4]) * 100,
			parts[5],
			2 if parts[10] == '.' else parts[10]])
	writer.close()

elif cnfile:
	logger.warning('All-FIT format detected, cnfile will be ignored!')

params.i = infile
params.d = outdir
params.o = outstem

shell.fg.allfit(**params)




from os import path, makedirs, remove
from shutil import rmtree, copyfile, move
from sys import stderr
from pyppl import Box
from bioprocs.utils import log2pyppl
from bioprocs.utils.shell2 import bgzip
import vcf

infile  = {{i.infile | quote}}
filters = {{args.filters | repr}}
fname   = {{args.filters | quote}}
gz      = {{args.gz | repr}}
outfile = {{o.outfile | quote}}
keep    = {{args.keep | repr}}
if gz:
	outfile = outfile[:-3]

"""
Builtin filters
"""
builtin_filters = Box(
	SNPONLY   = lambda r, s, q = None: len(r.REF) != 1 or any(len(a) != 1 for a in r.ALT),
	BIALTONLY = lambda r, s, q = None: len(r.ALT) != 1,
	QUAL      = lambda r, s, q: not r.QUAL or r.QUAL < q,
	AF        = lambda r, s, q: r.INFO['AF'][0] < q
)
builtin_descs = Box(
	SNPONLY   = lambda x: 'keep SNPs only',
	BIALTONLY = lambda x: 'keep bi-allelic mutations only',
	QUAL      = lambda x: 'keep sites with QUAL >= %s' % x,
	AF        = lambda x: 'keep sites with AF >= %s' % x
)
desc_prefix = 'Created by bioprocs.vcf.pVcfFilter: '

filters     = filters or {}
realfilters = {}
descs       = {}
for fname, ffunc in filters.items():
	if fname.startswith('!') and fname[1:] in builtin_filters:
		key = 'NOT' + fname[1:] + ['', str(ffunc)][int(bool(ffunc) or 0)]
		realfilters[key] = lambda r, s, fname = fname, ffunc = ffunc: \
			not builtin_filters[fname](r, s, ffunc) if ffunc not in [None, False] \
				else not builtin_filters[fname](r, s)
		descs[key] = desc_prefix + builtin_descs[fname[1:]](ffunc)
	elif fname in builtin_filters:
		key = [fname, fname + str(ffunc)][int(bool(ffunc) or 0)]
		realfilters[key] = lambda r, s, fname = fname, ffunc = ffunc: \
			builtin_filters[fname](r, s, ffunc) if ffunc not in [None, False] \
				else builtin_filters[fname](r, s)
		descs[key] = desc_prefix + builtin_descs[fname](ffunc)
	else:
		realfilters[fname] = ffunc if callable(ffunc) else eval(ffunc)
		descs[fname] = desc_prefix + fname

reader = vcf.Reader(filename=infile)
for fname, fdesc in descs.items():
	reader.filters[fname] = vcf.parser._Filter(id = fname, desc = fdesc)
writer = vcf.Writer(open(outfile, 'w'), reader)

while True:
	nerror = 0
	try:
		record = next(reader)
		for fname, ffunc in realfilters.items():
			if ffunc(record, record.samples):
				record.FILTER = record.FILTER or []
				record.FILTER.append(fname)
		if keep or not record.FILTER:
			writer.write_record(record)
	except StopIteration:
		break
	except Exception:
		nerror += 1
		import traceback
		tb = traceback.format_exc()
		log2pyppl(str(tb), 'ERROR')
		if nerror < 100:
			continue
		raise

writer.close()

if gz:
	bgzip(outfile)





from pyppl import Box
from bioprocs.utils import shell

infile   = {{ i.infile | quote}}
outfile  = {{ o.outfile | quote}}
bedtools = {{ args.bedtools | quote}}
params   = {{ args.params | repr }}
ref      = {{ args.ref | quote}}

shell.TOOLS.bedtools = bedtools

params.fi      = ref
params.bed     = infile
params._stdout = outfile
shell.Shell(subcmd = True, dash = '-', equal = ' ').bedtools.getfasta(**params).run()



#!/usr/bin/env python
"""Run DEG analysis for expression data"""

# Core information needed:
# 1. expression matrix, genes as rows and samples as columns, must be count data.
# 2. sample information file, specifying Samples, Patients (if paired analysis) and Group (comparison)

from os import path

from pyppl import PyPPL, Box
from bioprocs import params

params._prefix = '-'
params._desc = [
	"Things this script can do:",
	"1. Statistics/Characteristics of the expression matrix",
	"2. DEG Calling",
	"3. Real GSEA analysis",
	"4. Enrichment (depends on step 2)"]
params.exprmat.required = True
params.exprmat.desc     = 'The expression matrix.'
params.saminfo.required = True
params.saminfo.desc     = 'The sample information file.'
params.ppldir           = './workdir'
params.ppldir.desc      = 'The pipeline directory.'
params.exdir.required   = True
params.exdir.desc       = 'Where to export the result files.'
params.caller           = 'deseq2'
params.caller.desc      = 'The DEG caller'
params.runner           = 'local'
params.runner.desc      = 'The runner to run the processes.'
params.mapping          = ''
params.mapping.desc     = 'Probe to gene mapping'
params.nthread          = 1
params.nthread.desc     = '# threads to use.'
params.enlibs           = ['KEGG_2016']
params.enlibs.desc      = 'The enrichment analysis libraries'
params.cutoff           = 0.05
params.cutoff.desc      = 'The cutoff for DEGs'
params.cutoff.callback  = lambda opt: setattr(opt, 'value', {"by": "q", "value": opt.value} if not isinstance(opt.value, dict) else opt.value)
params.filter           = ""
params.filter.desc      = "A filter for the expression matrix"
params.filter.callback  = lambda opt: setattr(opt, 'value', opt.value or None)
params.steps            = ['stats', 'call', 'gsea', 'enrich']
params.steps.desc       = [
	'The steps to do with this script',
	'Use "-step" to clear defaults, and "-step <STEP>" to select step',
	'For example: "-step -step stats" will only run "stats",',
	'Only "-step stats" will not clear the default steps.'
]
params.steps.callback   = lambda opt: all(v in ['stats', 'call', 'gsea', 'enrich'] for v in opt.value) or "-steps should be a subset of ['stats', 'call', 'gsea', 'enrich']"

def main():
	opts = params._parse(dict_wrapper = Box)

	from bioprocs.gsea import pEnrichr
	from bioprocs.rnaseq import pExprStats, pRNASeqDEG

	# processes
	pExprStats.input = (opts.exprmat, opts.saminfo)
	pExprStats.args.tsform = 'function(x) log2(x+1)'
	pExprStats.args.opts.histogram.bins = 100

	pRNASeqDEG.args.tool    = opts.caller
	pRNASeqDEG.exdir        = path.join(opts.exdir, '2.DEGs')
	pRNASeqDEG.args.cutoff  = opts.cutoff
	pRNASeqDEG.args.mapfile = opts.mapping

	pEnrichr.args.genecol = 1 if opts.mapping else 0
	pEnrichr.args.libs = opts.enlibs
	pEnrichr.exdir = path.join(opts.exdir, '3.Enrichment', 'AllDEGs')

	if opts.filter:
		pExprStatsFiltered = pExprStats.copy()
		pExprStats.exdir = path.join(opts.exdir, '0.Unfiltered.expression.stats')
		pExprStatsFiltered.exdir = path.join(opts.exdir, '1.Filtered.expression.stats')
		pExprStatsFiltered.args.filter = opts.filter
	else:
		pExprStats.exdir = path.join(opts.exdir, '1.Expression.stats')

	# starts/depends
	starts = []
	if 'stats' in opts.steps:
		starts.append(pExprStats)
		if opts.filter:
			starts.append(pExprStatsFiltered)
	if 'call' in opts.steps:
		if 'stats' in opts.steps and opts.filter:
			pRNASeqDEG.depends = pExprStatsFiltered
			pRNASeqDEG.input = lambda ch: ch.expand(pattern = path.basename(opts.exprmat)).cbind(opts.saminfo)
		else:
			pRNASeqDEG.input = (opts.exprmat, opts.saminfo)
			starts.append(pRNASeqDEG)
	if 'gsea' in opts.steps:
		# to be implemented
		pass
	if 'enrich' in opts.steps:
		if 'call' not in opts.steps:
			raise ValueError('"enrich" step requires "call" step.')
		pEnrichr.depends             = pRNASeqDEG
		pEnrichr.args.pathview.fccol = int(opts.caller == 'deseq2') + int(bool(opts.mapping)) + 2
		pEnrichr.args.nthread        = opts.nthread
		pEnrichr.args.genecol        = int(bool(opts.mapping))
		pEnrichrUp                   = pEnrichr.copy()
		pEnrichrUp.exdir             = path.join(opts.exdir, '3.Enrichment', 'UpDEGs')
		pEnrichrUp.depends           = pRNASeqDEG
		pEnrichrUp.input             = lambda ch: ch.outdir.expand(pattern = '*.up.txt')
		pEnrichrDown                 = pEnrichr.copy()
		pEnrichrDown.exdir           = path.join(opts.exdir, '3.Enrichment', 'DownDEGs')
		pEnrichrDown.depends         = pRNASeqDEG
		pEnrichrDown.input           = lambda ch: ch.outdir.expand(pattern = '*.down.txt')

	config = {
		'default': {'ppldir': opts.ppldir, 'acache': True}
	}
	PyPPL(config).start(starts).run(opts.runner)

if __name__ == "__main__":
	main()






infiles = {{i.infiles | repr}}
outfile = {{o.outfile | quote}}
header  = {{args.header | repr}}

with open(infiles.pop(0)) as fin, open(outfile, 'w') as fout:
	fout.write(fin.read())

	for infile in infiles:
		with open(infile) as f:
			if header:
				f.readline()
			fout.write(f.read())
			


from pyppl import Box
from os import path
from bioprocs.utils import shell2 as shell
from bioprocs.utils.tsvio2 import TsvWriter
from gff import Gff

infile  = {{i.infile | quote}}
bedfile = {{i.bedfile | quote}}
bedext  = {{i.bedfile | ?:ext(_) == '.gz' | prefix | :_ | ext | [1:] | quote }}
outfile = {{o.outfile | quote}}
bwtool  = {{args.bwtool | quote}}
params  = {{args.params | repr}}

if 'bed' not in bedext:
	def getname(record):
		for key, val in record['attributes'].items():
			if key.endswith('id'):
				return val
		for key, val in record['attributes'].items():
			if key.endswith('name'):
				return val
		return f'{record["seqid"]}:{record["start"]}-{record["end"]}'

	tmpfile = path.join(path.dirname(outfile), 'regions.bed')
	gff = Gff(bedfile)
	writer = TsvWriter(tmpfile)
	for r in gff:
		writer.write([r['seqid'], r['start'], r['end'], getname(r)])
	writer.close()
	bedfile = tmpfile

shell.load_config(bwtool = dict(_exe = bwtool, _prefix = '-'))

shell.fg.bwtool.summary(bedfile, infile, outfile, **params)



"""Power analysis"""
from pyppl import Proc, Box
from . import params, rimport
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pSurvivalPower():
	"""
	@name:
		pSurvivalPower
	@description:
		Do power analysis for survival analysis.
		See http://www.sample-size.net/sample-size-survival-analysis/
	@input:
		`infile:file`: The input file, could be either:
			- detailed suvival data with [`patient`, ]`time`, `status`, `variable1`, `variable2`, ...; or
			- ratios with `variable`, `survrate1`, `survrate2`, `ssratio`, where `survrate1` and
				`survrate2` are survival rates in group1 and group2, respectively,
				and `ssratio` is sample size ratio in group1/group2
		`ngroup1`: The size of 1st group, for detailed input file
		`ngroup2`: The size of 2nd group, for detailed input file
		`ngroup3`: The size of 3rd group, for detailed input file
		`ngroup4`: The size of 4th group, for detailed input file
	@output:
		`outfile:file`: The output file with columns:
			- Variable: the variable (including paired groups)
			- Alpha: the alpha value
			- Beta: the beta value (1-power)
			- SSize1: the sample size for group1
			- SSize2: the sample size for group2
			- Total: the total sample size
	@args:
		`rnames`: Whether the detailed input file has row names. Default: `True`
		`plot`  : Plot the results? Default: `False`
		`ngroup`: Number of groups to divide into for detailed input file. Default: `2`
		`intype`: The input file type. Either `detailed` (default) or `ratio`
		`alphas`: The alpha values (two-sided). You need to *2 for one-sided. Default:
			- `[.005, .01, .05, .1]`
		`betas` : 1-power. Default: `[.05, .1, .2]`
	"""
	pSurvivalPower              = Proc(desc = "Survival analysis.")
	pSurvivalPower.input        = 'infile:file, ngroup1, ngroup2, ngroup3, ngroup4'
	pSurvivalPower.output       = 'outfile:file:{{i.infile | fn}}.power/{{i.infile | fn}}.ssize.txt, outdir:dir:{{i.infile | fn}}.power'
	pSurvivalPower.args.rnames  = True
	pSurvivalPower.args.plot    = False
	pSurvivalPower.args.ngroup  = 2
	pSurvivalPower.args.intype  = 'detailed' # ratios or ratio
	pSurvivalPower.args.alphas  = [.005, .01, .05, .1] # two-sided, *2 if you want one-sided
	pSurvivalPower.args.betas   = [.05, .1, .2]
	pSurvivalPower.envs.rimport = rimport
	pSurvivalPower.lang         = params.Rscript.value
	pSurvivalPower.script       = "file:scripts/power/pSurvivalPower.r"
	return pSurvivalPower




from pyppl import Box
from pyppl.utils import cmd
from collections import OrderedDict
from subprocess import list2cmdline

try:  # py3
	from shlex import quote
except ImportError:  # py2
	from pipes import quote

shquote = lambda x: x[6:] if str(x).startswith('shraw:') else quote(str(x))

TOOLS = Box()

class RuncmdException(Exception):
	pass

def cmdargs(params, dash = 'auto', equal = 'auto', duplistkey = False, ignorefalse = True):
	"""
	Convert a dict of parameters into a command line parameter
	@params:
		`params`: The dict of parameters
			- positional arguments will be with key "_"
			- positional arguments before all other arguments will be with key ""
			- argument with key "_stdout" will redirct the stdout to file:
				- {'_stdout': 'outfile'} => " > 'outfile'"
		`dash`  : The dash prefix of each parameter key.
			- `auto`: A single dash for single-char key, otherwise double dashes
			- Otherwise use whatever being passed.
		`equal` : The separater between the key and the value.
			- `auto`: Space for single-char key, otherwise equal sign `=`
			- Otherwise use whatever being passed.
		`duplistkey`: Whether duplicate keys for list values. Default: `False`
			- For example for `params = {'a': [1,2,3]}`,
			- if `duplistkey = True`, it will be: `-a 1 -a 2 -a 3`,
			- otherwise it will be `-a 1 2 3`
		`ignorefalse`: Ignore boolean False parameter? Default: `True`
	"""
	if not params: return ''

	ret = []
	# decide the order
	params2 = OrderedDict()
	if '' in params:
		params2['']  = params['']
		del params['']
	if isinstance(params, OrderedDict):
		for key, val in params.items():
			if key == '_':
				continue
			params2[key] = val
		if '_' in params:
			params2['_'] = params['_']
	else:
		for key in sorted(params.keys()):
			if key == '_':
				continue
			params2[key] = params[key]
		if '_' in params:
			params2['_'] = params['_']

	params = params2
	outfile = None
	if '_stdout' in params:
		outfile = params['_stdout']
		del params['_stdout']
	aoutfile = None
	if '__stdout' in params:
		aoutfile = params['__stdout']
		del params['__stdout']
	elif '_stdout_' in params:
		aoutfile = params['_stdout_']
		del params['_stdout_']
	if outfile and aoutfile:
		raise ValueError('Cannot have both out files to write and append to.')
	for key, val in params.items():
		if key in ['', '_']:
			if not isinstance(val, (list, tuple)):
				val = [val]
			ret.extend([shquote(v) for v in val])
			continue
		# allow comments for parameters for the same key
		key = key.split('#')[0].strip()
		item = dash if dash != 'auto' else '--' if len(key) > 1 else '-'
		item += key
		if isinstance(val, (tuple, list)):
			# ignore keys with no values
			if not val: continue
			item += equal if equal != 'auto' else '=' if len(key)>1 else ' '
			if duplistkey:
				ret.extend([item + shquote(v) for v in val])
			else:
				ret.append(item + shquote(str(val[0])))
				ret.extend([shquote(str(v)) for v in val[1:]])
		elif isinstance(val, bool):
			if not val and ignorefalse:
				continue
			if val:
				ret.append(item)
			else:
				item += equal if equal != 'auto' else '=' if len(key)>1 else ' '
				item += '0'
				ret.append(item)
		else:
			item += equal if equal != 'auto' else '=' if len(key)>1 else ' '
			item += shquote(val)
			ret.append(item)

	if outfile:
		ret.append('>')
		ret.append(shquote(outfile))
	elif aoutfile:
		ret.append('>>')
		ret.append(shquote(aoutfile))
	return ' '.join(ret)

def runcmd(cmd2run, raiseExc = True, **kwargs):
	from . import logger
	logger.info('RUNNING: %s', list2cmdline(cmd2run) if isinstance(cmd2run, list) else cmd2run)
	# stdin to avoid blocking
	kwargs['shell'] = kwargs.get('shell', True)
	c = cmd.Cmd(cmd2run, stdin = None, **kwargs)
	logger.info('- PID: %s', c.pid)

	for line in c.readline():
		logger.info('- STDERR: %s', line.rstrip())

	logger.info('- RETURNCODE: %s', c.rc)
	logger.info('-' * 80)
	if raiseExc and c.rc != 0:
		raise RuncmdException(c.cmd)
	return c

def runcmd2(tool, params, dash = 'auto', equal = 'auto', duplistkey = False, ignorefalse = True, quit = True):
	s = Shell(tools = {'tool': tool}, dash = dash, equal = equal, duplistkey = duplistkey, ignorefalse = ignorefalse)
	r = s.tool(**params).run()
	if quit and r.rc != 0:
		raise RuncmdException(r.cmd)
	return r

# def call(command, args, quit = True):
# 	bioprocs = path.join(path.realpath(path.dirname(path.dirname(path.dirname(__file__)))), 'bin', 'bioprocs')
# 	if not isinstance(args, list):
# 		args = shlex.split(args)
# 	args = [bioprocs, command] + args
# 	return runcmd(args, quit)

class ShellResult(object):

	def __init__(self, cmdobj, tools = None):
		self.cmdobj = cmdobj
		self.tools  = tools or {}
		self.done   = False
	
	def __str__(self):
		return self.stdout

	def __repr__(self):
		return '<ShellResult {!r}>'.format(self.cmdobj)

	def run(self, raiseExc = True, save = None, logger = 'bioprocs'):
		if not self.done:
			if logger == 'bioprocs':
				from . import logger
				logit = lambda *args: logger.info(*args)
			elif logger == 'sys':
				from sys import stderr
				logit = lambda *args: stderr.write((args[0] + '\n') % args[1:])
			else:
				logit = lambda *args: None
			logit('RUNNING: %s', self.cmdobj.cmd)
			logit('- PID: %s', self.cmdobj.pid)

			for line in self.cmdobj.readline(save = save):
				logit('- STDERR: %s', line.rstrip())
			self.done = True
			logit('- RETURNCODE: %s', self.cmdobj.rc)
			logit('-' * 80)
			if raiseExc and self.cmdobj.rc != 0:
				raise RuncmdException(self.cmdobj.cmd)
		return self

	@property
	def stdout(self):
		return self.run(save = 'stdout').cmdobj.stdout
	
	@property
	def stderr(self):
		return self.run(save = 'stderr').cmdobj.stderr

	@property
	def cmd(self):
		return self.cmdobj.cmd

	@property
	def returncode(self):
		return self.run().cmdobj.rc
	
	@property
	def rc(self):
		return self.returncode

	def pipe(self, 
		tools = None,   subcmd = False,
		dash  = 'auto', equal  = 'auto', duplistkey = False, ignorefalse = True,
		base  = None,   **pargs):
		tools = tools or {}
		tools.update(self.tools)
		return Shell(
			tools       = tools,
			subcmd      = subcmd,
			dash        = dash,
			equal       = equal,
			duplistkey  = duplistkey,
			ignorefalse = ignorefalse,
			base        = None,
			stdin       = self.cmdobj.p.stdout,
			prevcmd     = self.cmd,
			**pargs
		)

class Shell(object):
	def __init__(self, 
		tools = None,   subcmd  = False,
		dash  = 'auto', equal   = 'auto', duplistkey = False, ignorefalse = True,
		base  = None,   prevcmd = None,   **pargs):
		self.dash        = dash
		self.equal       = equal
		self.duplistkey  = duplistkey
		self.ignorefalse = ignorefalse
		self.subcmd      = subcmd
		self.tools       = TOOLS
		self.base        = base
		self.prevcmd     = prevcmd

		self.tools.update(tools or {})
		# common args for tools with subcommand
		self.targs         = {}
		self._subcmdCalled = False

		self.pargs = pargs.copy()
		self.pargs['stdin'] = self.pargs.get('stdin')
		self.pargs['shell'] = True

	def __call__(self, *args, **kwargs):
		self.targs.update(kwargs)
		self.targs[''] = list(args) + self.targs.get('', [])
		return self

	def _run(self, *args, **kwargs):
		kwargs[''] = list(args) + kwargs.get('', []) 
		targs = '' if not self.targs else ' ' + cmdargs(
			self.targs, dash = self.dash, equal = self.equal, 
			duplistkey = self.duplistkey, ignorefalse = self.ignorefalse
		)
		cargs = cmdargs(
			kwargs,     dash = self.dash, equal = self.equal, 
			duplistkey = self.duplistkey, ignorefalse = self.ignorefalse
		)
		cmd2run = '{0}{1} {2}'.format(self.base, targs, cargs)
		cmdobj  = cmd.Cmd(cmd2run, **self.pargs)
		
		if self.prevcmd:
			cmdobj.cmd = '{} | {}'.format(self.prevcmd, cmdobj.cmd)
		return ShellResult(cmdobj, self.tools)

	def __getitem__(self, tool):
		return getattr(self, tool)

	def __getattr__(self, tool):
		tool = self.tools.get(tool, tool)
		# there is not subcommand or now it's already in subcommand
		if not self.subcmd:
			# in a subcommand
			if self.base:
				if not self._subcmdCalled:
					self.targs['_'] = self.targs.get('_', []) + [tool]
					self._subcmdCalled = True
				else:
					self.targs['_'].pop(-1)
					self.targs['_'].append(tool)
			# plain command
			else:
				self.base = shquote(tool)
			return self._run
		return Shell(
			subcmd = False, dash = self.dash, equal = self.equal, duplistkey = self.duplistkey, 
			ignorefalse = self.ignorefalse, base = tool, **self.pargs)

wc   = lambda *args, **kwargs: Shell().wc(*args, **kwargs).run(save = 'stdout')
wc_l = lambda filename: int(wc(l = filename).stdout.split()[0])
wcl  = wc_l

gzip      = lambda *args,  **kwargs: Shell().gzip(*args, **kwargs).run(logger = False)
bgzip     = lambda *args,  **kwargs: Shell(equal = ' ').bgzip(*args, **kwargs).run(logger = False)
gunzip    = lambda *args,  **kwargs: Shell().gunzip(*args, **kwargs).run(logger = False)
gzip_to   = lambda source, dest: gzip(_ = source, c = True, _stdout = dest)
bgzip_to  = lambda source, dest: bgzip(_ = source, c = True, _stdout = dest)
gunzip_to = lambda source, dest: gunzip(_ = source, c = True, _stdout = dest)

mv    = lambda *args,  **kwargs: Shell().mv(*args, **kwargs).run(logger = False)
cp    = lambda *args,  **kwargs: Shell().cp(*args, **kwargs).run(logger = False)
rm    = lambda *args,  **kwargs: Shell().rm(*args, **kwargs).run(logger = False)
rm_rf = lambda *args: rm(*args, r = True, f = True)
rmrf  = rm_rf
mkdir = lambda *args,  **kwargs: Shell().mkdir(*args, **kwargs).run(logger = False)
ln    = lambda *args,  **kwargs: Shell().ln(*args, **kwargs).run(logger = False)
ln_s  = lambda source, dest: ln(source, dest, s = True)

head  = lambda *args, **kwargs: Shell().head(*args, **kwargs).run(logger = False)
tail  = lambda *args, **kwargs: Shell().tail(*args, **kwargs).run(logger = False)
grep  = lambda *args, **kwargs: Shell().grep(*args, **kwargs).run(raiseExc = False, logger = False)
sort  = lambda *args, **kwargs: Shell(dash = '-', equal = ' ', duplistkey = True).sort(*args, **kwargs).run(logger = False)
uniq  = lambda *args, **kwargs: Shell().uniq(*args, **kwargs).run(logger = False)
touch = lambda *args, **kwargs: Shell().touch(*args, **kwargs).run(logger = False)

kill  = lambda *args, **kwargs: Shell().kill(*args, **kwargs).run(logger = False)
kill9 = lambda *procids: kill(**{'9': True, '': procids})

zcat  = lambda *args,  **kwargs: Shell().zcat(*args, **kwargs).run(logger = False)
cat   = lambda *args,  **kwargs: Shell().cat(*args, **kwargs).run(logger = False)
# _ means waiting for run
acat_ = lambda infile, **kwargs: Shell().zcat(infile, **kwargs) if infile.endswith('.gz') else Shell().cat(infile, **kwargs)

which = lambda *args, **kwargs: Shell().which(*args, **kwargs).run(logger = False, save = 'stdout').stdout.rstrip('\n')



import sys

from os import path
from pyppl import Box
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter
from bioprocs.utils import log2pyppl

infile   = {{i.infile | quote}}
outfile  = {{o.outfile | quote}}
region   = {{args.region | repr}}
notfound = {{args.notfound | quote}}
inopts   = {{args.inopts | repr}}
refgene  = {{args.refgene | quote}}
genecol  = {{args.genecol | repr}}
if region.down is None:
	region.down = region.up

# get all genes' TSS and strand
reader = TsvReader(refgene, cnames = False, delimit = '"')
genes  = {r[1]:r[0].split("\t")[:7] for r in reader}

reader = TsvReader(infile, **inopts)
writer = TsvWriter(outfile)
for r in reader:
	gene = r[genecol]
	if gene not in genes:
		msg = 'Gene does not exist: {}'.format(gene)
		if notfound == 'error':
			raise ValueError(msg)
		else:
			log2pyppl('Gene does not exist: {msg}', 'warning')
			continue
	chrom, _, _, start, end, _, strand = genes[gene]
	start, end = int(start), int(end)
	if strand == '-':
		record = [
			chrom, 
			max(0, min(start, end - region.down) if region.withbody else end - region.down), 
			end + region.up, 
			gene, 0, strand
		]
	else:
		record = [
			chrom, 
			max(0, start - region.up), 
			max(end, start + region.down) if region.withbody else start + region.down, 
			gene, 0, strand
		]
	writer.write(record)
writer.close()




import math
from pyppl import Box
from functools import partial
from bioprocs.utils import alwaysList
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

infile  = {{ i.infile | quote}}
outfile = {{ o.outfile | quote}}
inopts  = {{ args.inopts | repr}}
on      = {{ args.on if isinstance(args.on, int) or args.on.startswith('lambda') else quote(args.on) }}
helper  = {{args.helper | repr}}
if not isinstance(helper, list):
	helper = [helper]

# built-in aggregators
def aggr_sum(rs, idx):
	ret = [0] * len(idx)
	for r in rs:
		for i, ix in enumerate(idx):
			ret[i] += float(r[ix])
	return ret[0] if len(idx) == 1 else ret

def aggr_mean(rs, idx):
	lenrs = float(len(rs))
	sums  = aggr_sum(rs, idx)
	if len(idx) == 1:
		return sums/lenrs
	return [s/lenrs for s in sums]

def get_median(ls):
	ls = list(sorted(float(l) for l in ls))
	lenls = len(ls)
	if lenls % 2 == 1:
		return ls[int((lenls-1)/2.0)]
	return (ls[int((lenls-2)/2.0)] + ls[int(lenls/2.0)])/2.0

def aggr_median(rs, idx):
	if len(idx) == 1:
		return get_median(r[idx[0]] for r in rs)
	lss = [[] for _ in idx]
	for r in rs:
		for i, ix in enumerate(idx):
			lss[i].append(r[ix])
	return [get_median(ls) for ls in lss]

def aggr_min(rs, idx):
	if len(idx) == 1:
		return min(float(r[idx[0]]) for r in rs)
	ret = [None] * len(idx)
	for r in rs:
		for i, ix in enumerate(idx):
			ret[i] = r[ix] if ret[i] is None else min(ret[i], float(r[ix]))
	return ret

def aggr_max(rs, idx):
	if len(idx) == 1:
		return max(float(r[idx[0]]) for r in rs)
	ret = [None] * len(idx)
	for r in rs:
		for i, ix in enumerate(idx):
			ret[i] = r[ix] if ret[i] is None else max(ret[i], float(r[ix]))
	return ret

def aggr_fisher(rs, idx):
	from scipy.stats import combine_pvalues
	if (len(idx) == 1):
		return combine_pvalues([float(r[idx[0]]) for r in rs])[1]
	ret = [None] * len(idx)
	for i, ix in enumerate(idx):
		ret[i] = combine_pvalues([float(r[ix]) for r in rs])[1]
	return ret

builtin = {
	"sum"   : aggr_sum,
	"mean"  : aggr_mean,
	"median": aggr_median,
	"min"   : aggr_min,
	"max"   : aggr_max,
	"fisher": aggr_fisher,
}

helper = [line for line in helper if line]
exec('\n'.join(helper), globals())
aggrs  = Box()
naggrs = {}
{% for col, func in args.aggrs.items() %}
col  = {{col | quote}}
aggrs [col] = {{func if not func.startswith('$') else quote(func)}}
if not callable(aggrs[col]) and not aggrs[col].startswith('$'):
	raise ValueError("I don't know how for {!r}.".format(col))
if not callable(aggrs[col]):
	fn, args = aggrs[col].split(':', 1)
	fn = fn.strip()[1:]
	if fn not in builtin:
		raise ValueError('Unknown builtin aggregation function, expect sum, mean, median, min or max.')
	args = [int(arg) if arg.isdigit() else arg for arg in args.strip().split(',')]
	aggrs[col] = partial(builtin[fn], idx = args)
naggrs[col] = len(alwaysList(col))
{% endfor %}


reader = TsvReader(infile, **inopts)
writer = TsvWriter(outfile, delimit = inopts.get('delimit', "\t"))
if reader.cnames:
	if isinstance(on, int):
		writer.cnames = [reader.cnames[on]] + alwaysList(','.join(aggrs.keys()))
	elif callable(on):
		writer.cnames = ['AggrGroup'] + alwaysList(','.join(aggrs.keys()))
	elif on not in reader.cnames:
		raise ValueError('{!r} is not a valid column name!'.format(on))
	else:
		writer.cnames = [on] + alwaysList(','.join(aggrs.keys()))
		on = reader.cnames.index(on)
elif callable(on):
	writer.cnames = ['AggrGroup'] + alwaysList(','.join(aggrs.keys()))
elif not isinstance(on, int):
	raise ValueError('The input file does not have column name (args.inopts.cnames = False?).')
else:
	writer.cnames = ['ROWNAME'] + alwaysList(','.join(aggrs.keys()))

writer.writeHead()

refcol = None
rows   = []
for row in reader:
	if callable(on):
		col = on(row)
	else:
		col = row[on]
	if refcol is None:
		refcol = col
		rows.append(row)
	elif col == refcol:
		rows.append(row)
	else:
		out = [refcol]
		for c, a in aggrs.items():
			if naggrs[c] > 1:
				out.extend(a(rows))
			else:
				out.append(a(rows))
		writer.write(out)
		refcol = col
		rows   = [row]
out = [refcol]
for c, a in aggrs.items():
	if naggrs[c] > 1:
		out.extend(a(rows))
	else:
		out.append(a(rows))
writer.write(out)
writer.close()






from pyppl import Box
from bioprocs.utils import regionOverlap
from bioprocs.utils.tsvio import TsvReader, TsvWriter

# read input file
regfile = {{i.regfile | quote}}
outfile = {{o.outfile | quote}}
ext     = {{i.regfile | ext | [1:] | quote}}
regopts = {{args.regopts}}
regions = None

if regopts.ftype == 'auto':
	regopts.ftype = ext
reader  = TsvReader(regfile, **regopts)
regions = reader.dump()

intfile = {{i.intfile | quote}}
ext     = {{i.intfile | ext | [1:] | quote}}
intopts = {{args.intopts}}

writer  = TsvWriter(outfile, ftype = 'bed')
writer.meta.add('CHR2', 'START2', 'END2', 'INTNAME')
writer.writeHead()

if intopts.ftype == 'auto':
	intopts.ftype = ext

import re
reader = TsvReader(intfile, **intopts)
if intopts.ftype == 'bedx':
	reader.meta.add('CHR2', 'START2', 'END2')

for line in reader:
	if intopts.ftype == 'bedpe':
		reg1 = [line.CHR1, line.START1, line.END1]
	else:
		reg1 = [line.CHR, line.START, line.END]
	if intopts.ftype == 'bed12':
		reg2 = re.split(r'[^\w]+', line.NAME)[:3]
	else: 
		reg2 = [line.CHR2, line.START2, line.END2]

	for region in regions:
		if regionOverlap(*(reg1 + [region.CHR, region.START, region.END])):
			region.CHR2   = reg2[0]
			region.START2 = reg2[1]
			region.END2   = reg2[2]
			region.INTNAME= line.NAME
			writer.write(region)
		if regionOverlap(*(reg2 + [region.CHR, region.START, region.END])):
			region.CHR2   = reg1[0]
			region.START2 = reg1[1]
			region.END2   = reg1[2]
			region.INTNAME= line.NAME
			writer.write(region)



"""Analysis of expression data from RNA-seq"""
from os import path
from glob import glob
from pyppl import Proc, Box
from . import params, rimport
from .utils import dirpat2name, fs2name
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pExprDir2Matrix():
	"""
	@name:
		pExprDir2Matrix
	@description:
		Convert expression files to expression matrix
		File names will be used as sample names (colnames)
		Each gene and its expression per line.
		Suppose each expression file has the same rownames and in the same order.
	@input:
		`indir:file`:  the directory containing the expression files, could be gzipped
	@output:
		`outfile:file`: the expression matrix file
		`outdir:dir`:   the directory containing expr file and plots
	@args:
		`pattern` : The pattern to filter files. Default `'*'`
		`namefunc`: Transform filename (no extension) as column name. Default: "function(fn) fn"
		`header`  : Whether each expression file contains header. Default: `False`
		`exrows`  : Rows to be excluded, regular expression applied. Default: `["^Sample", "^Composite", "^__"]`
		`boxplot` : Whether to plot a boxplot. Default: False
		`heatmap` : Whether to plot a heatmap. Default: False
		`histplot`: Whether to plot a histgram. Default: False
		`devpars` : Parameters for png. Default: `{'res': 300, 'width': 2000, 'height': 2000}`
		`boxplotggs`: The ggplot parameters for boxplot. Default: `['r:ylab("Expression")']`
			- See ggplot2 documentation.
		`heatmapggs`: The ggplot parameters for heatmap. Default: `['r:theme(axis.text.y = element_blank())']`
		`histplotggs`: The ggplot parameters for histgram. Default: `['r:labs(x = "Expression", y = "# Samples")']`
	"""
	pExprDir2Matrix        = Proc(desc = 'Merge expression files to a matrix.')
	pExprDir2Matrix.input  = "indir:file"
	pExprDir2Matrix.output = [
		"outfile:file:{{i.indir, args.pattern | dirpat2name}}.dir/{{i.indir, args.pattern | dirpat2name}}.expr.txt",
		"outdir:dir:{{i.indir, args.pattern | dirpat2name}}.dir"
	]
	pExprDir2Matrix.lang           = params.Rscript.value
	pExprDir2Matrix.args.pattern   = '*'
	pExprDir2Matrix.args.fn2sample = 'function(fn) fn'
	pExprDir2Matrix.args.exrows    = ["^Sample", "^Composite", "^__"]
	pExprDir2Matrix.args.hmrows    = 500
	pExprDir2Matrix.args.plot = Box(
		boxplot = False,
		heatmap = False,
		histogram = False
	)
	pExprDir2Matrix.args.ggs = Box(
		boxplot = Box(ylab = {0: "Expression"}),
		heatmap = Box(theme = {'axis.text.y': 'r:element_blank()'}),
		histogram = Box(labs = {'x': 'Expression', 'y': '# Genes'})
	)
	pExprDir2Matrix.envs.dirpat2name = dirpat2name
	pExprDir2Matrix.envs.rimport = rimport
	pExprDir2Matrix.args.devpars = Box(res = 300, width = 2000, height = 2000)
	pExprDir2Matrix.script       = "file:scripts/rnaseq/pExprDir2Matrix.r"
	return pExprDir2Matrix

@procfactory
def _pExprFiles2Mat():
	"""
	@name:
		pExprFiles2Mat
	@description:
		Merge expression to a matrix from single samples.
	@input:
		`infiles:files`: The expression file from single samples, typically with 2 columns: gene and expression value
	@output:
		`outfile:file`: the expression matrix file
	@args:
		`fn2sample`: Transform filename (no extension) as column name. Default: "function(fn) fn"
		`inopts`   : Options to read input files. Default: `Box(rname = True, cnames = True)`
	"""
	pExprFiles2Mat                = Proc(desc = 'Merge expression to a matrix from single samples.')
	pExprFiles2Mat.input          = 'infiles:files'
	pExprFiles2Mat.output         = 'outfile:file:{{i.infiles | fs2name}}.expr.txt'
	pExprFiles2Mat.args.inopts    = Box(cnames = True, rnames = True)
	pExprFiles2Mat.args.fn2sample = 'function(fn) unlist(strsplit(fn, ".", fixed=T))[1]'
	pExprFiles2Mat.envs.fs2name   = fs2name
	pExprFiles2Mat.envs.rimport   = rimport
	pExprFiles2Mat.lang           = params.Rscript.value
	pExprFiles2Mat.script         = "file:scripts/rnaseq/pExprFiles2Mat.r"
	return pExprFiles2Mat

@procfactory
def _pExprStats():
	"""
	@name:
		pExprStats
	@description:
		Plot the expression out.
	@input:
		`infile:file`: The expression matrix (rows are genes and columns are samples).
		`gfile:file` : The sample information file. Determines whether to do subgroup stats or not.
			- If not provided, not do for all samples
	@output:
		`outdir:dir`: The directory containing the plots
			- If `args.filter` is given, a filtered expression matrix will be generated in `outdir`.
	@args:
		`inopts`: Options to read `infile`. Default: `Box(cnames = True, rnames = True)`
		`tsform`: An R function in string to transform the expression matrix (i.e take log).
		`filter`: An R function in string to filter the expression data.
		`plot`  : Which plot to do? Default: `Box(boxplot = True, histogram = True, qqplot = True)`
		`ggs`   : The ggs for each plot. Default:
			- `boxplot   = Box(ylab = {0: "Expression"})`,
			- `histogram = Box(labs = {'x': 'Expression', 'y': '# Genes'})`,
			- `qqplot    = Box()`
		`params` : The params for each ggplot function.
		`devpars`: Parameters for png. Default: `{'res': 300, 'width': 2000, 'height': 2000}`
	"""
	pExprStats             = Proc(desc = 'Plot the expression values out.')
	pExprStats.input       = 'infile:file, gfile:file'
	pExprStats.output      = 'outdir:dir:{{i.infile | fn2}}.plots'
	pExprStats.args.inopts = Box(cnames = True, rnames = True)
	pExprStats.args.tsform = None
	pExprStats.args.filter = None
	pExprStats.args.plot   = Box(boxplot = True, histogram = True, qqplot = True)
	pExprStats.args.ggs    = Box(
		boxplot   = Box(ylab = {0: "Expression"}),
		histogram = Box(labs = {'x': 'Expression', 'y': '# Genes'}),
		qqplot    = Box()
	)
	pExprStats.args.params   = Box(
		boxplot   = Box(),
		histogram = Box(),
		qqplot    = Box()
	)
	pExprStats.args.devpars = Box(res = 300, width = 2000, height = 2000)
	pExprStats.envs.rimport = rimport
	pExprStats.lang         = params.Rscript.value
	pExprStats.script       = "file:scripts/rnaseq/pExprStats.r"
	return pExprStats

@procfactory
def _pBatchEffect():
	"""
	@name:
		pBatchEffect
	@description:
		Remove batch effect with sva-combat.
	@input:
		`expr:file`:  The expression file, generated by pExprDir2Matrix
		`batch:file`: The batch file defines samples and batches.
	@output:
		`outfile:file`: the expression matrix file
		`outdir:dir`:   the directory containing expr file and plots
	@args:
		`tool`    : The tool used to remove batch effect. Default `'combat'`
		`hmrows`  : How many rows to be used to plot heatmap
		`plot`: Whether to plot
			- `boxplot`   : Whether to plot a boxplot. Default: False
			- `heatmap`   : Whether to plot a heatmap. Default: False
			- `histogram` : Whether to plot a histgram. Default: False
		`devpars`    : Parameters for png. Default: `{'res': 300, 'width': 2000, 'height': 2000}`
		`ggs`: The ggplot parameters
			- `boxplot`  : The ggplot parameters for boxplot. Default: `Box(ylab = {0: "Log2 Intensity"})`
			- `heatmap`  : The ggplot parameters for heatmap. Default: `Box(theme = {'axis.text.y': 'r:element_blank()'})`
			- `histogram`: The ggplot parameters for histgram. Default: `Box(labs = {'x': "Log2 Intensity", "y": "Density"})`
	"""
	pBatchEffect              = Proc(desc = 'Try to remove batch effect of expression data.')
	pBatchEffect.input        = "expr:file, batch:file"
	pBatchEffect.output       = "outfile:file:{{i.expr | fn2}}/{{i.expr | fn2}}.expr.txt, outdir:dir:{{i.expr | fn2}}"
	pBatchEffect.args.tool    = 'combat'
	pBatchEffect.args.plot    = Box(boxplot = False, heatmap = False, histogram = False)
	pBatchEffect.args.hmrows  = 500
	pBatchEffect.args.devpars = Box(res = 300, width = 2000, height = 2000)
	pBatchEffect.args.ggs     = Box(
		boxplot   = Box(ylab = {0: "Expression"}),
		heatmap   = Box(theme = {'axis.text.y': 'r:element_blank()'}),
		histogram = Box(labs  = {'x': "Expression", "y": "Frequency"})
	)
	pBatchEffect.envs.rimport = rimport
	pBatchEffect.lang   = params.Rscript.value
	pBatchEffect.script = "file:scripts/rnaseq/pBatchEffect.r"
	return pBatchEffect

@procfactory
def _pUnitConversion():
	"""
	@description:
		Convert expression between units
		See here: https://haroldpimentel.wordpress.com/2014/05/08/what-the-fpkm-a-review-rna-seq-expression-units/ and
		https://docs.gdc.cancer.gov/Data/Bioinformatics_Pipelines/Expression_mRNA_Pipeline/#fpkm
		Available converstions:
		- `count -> cpm, fpkm/rpkm, fpkm-uq/rpkm-rq, tpm, tmm`
		- `fpkm/rpkm -> count, tpm, cpm`
		- `tpm -> count, fpkm/rpkm, cpm`
		- `cpm -> count, fpkm/rpkm, tpm`
		NOTE that during some conversions, `sum(counts/effLen)` is approximated to `sum(counts)/sum(effLen) * length(effLen))`
	@input:
		`infile`: the expression matrix
			- rows are genes, columns are samples
	@output:
		`outfile`: the converted expression matrix
	@args:
		`inunit` : The unit of input expression values.
		`outunit`: The unit of output expression values.
		`meanfl` : A file containing the mean fragment length for each sample by rows, without header.
			- Or a fixed universal estimated number (1 used by TCGA).
		`nreads`:    The estimatied total number of reads for each sample.
			- Or you can pass a file with the number for each sample by rows, without header.
			- In converting `fpkm/rpkm -> count`: it should be total reads of that sample
			- In converting `cpm -> count`: it should be total reads of that sample
			- In converting `tpm -> count`: it should be total reads of that sample
			- In converting `tpm -> cpm`: it should be total reads of that sample
			- In converting `tpm -> fpkm/rpkm`: it should be `sum(fpkm)` of that sample
		`inform`:    Transform the input to the unit specified. (sometimes the values are log transformed)
			- For example, if the `inunit` is `tpm`, but it's actually `log2(expr+1)` transformed,
			- Then this should be `function(expr) 2^(expr - 1)`.
		`outform`: Transform to be done on the output expression values.
		`refexon`: the exome gff file, for RPKM/FPKM
			- `gene_id` is required for gene names
	@requires:
		[edgeR](https://bioconductor.org/packages/release/bioc/html/edger.html) if cpm or rpkm is chosen
		[coseq](https://rdrr.io/rforge/coseq/man/transform_RNAseq.html) if tmm is chosen
	"""
	return Box(
		desc   = 'Convert RNAseq data in different units back and forth',
		lang   = params.Rscript.value,
		input  = 'infile:file',
		output = "outfile:file:{{i.infile | fn2}}.{{args.outunit}}{{args.outform | lambda x: '_t' if x else ''}}.txt",
		args   = Box(
			inunit  = 'count',
			outunit = 'tpm',
			meanfl  = 1,
			nreads  = 50000000,
			refexon = params.refexon.value,
			inform  = None,
			outform = None
		)
	)

@procfactory
def _pRNASeqDEG():
	"""
	@input:
		efile:file: The expression matrix
			- Columns other than samples in gfile will be used as annotations
			- See `args.mapping`
		gfile:file: The group information
			- For example:
				```
				Sample	[Patient	]Group
				sample1	[patient1	]group1
				sample2	[patient1	]group1
				sample3	[patient2	]group1
				sample4	[patient2	]group2
				sample5	[patient3	]group2
				sample6	[patient3	]group2
				```
			- If it is not paired comparison, you should not include Patient.
	@output:
		outfile:file: The DEG list
		outdir:file:  The output directory containing deg list and plots
	@args:
		tool  : The tool used to detect DEGs. Default: 'deseq2' (edger is also available).
		inopts: Options to read `infile`. Default: `Box(cnames = True, rnames = True)`
		cutoff: The cutoff used to filter the results. Default: `0.05`
			- `0.05` implies `{"by": "p", "value": "0.05", "sign": "<"}`
		ggs   : The ggs for each plot. Default:
			- For heatmap: should be the `draw` argument from `plot.heatmap2` in `plot.r`
			- Not available for `mdsplot`.
			- Others are empty `Box()`s
			- To disable a plot: `ggs.heatmap = FALSE`
		params: Parameters for each plot. Default:
			- `volplot`: `Box(pcut = 0.05, logfccut = 2)`
			- `maplot` : `Box(pcut = 0.05)`
			- `heatmap`: `Box(ngenes = None, <other arguments for plot.heatmap2's params>)`, all genes in heatmap or a number for up/down genes in heatmap
		devpars: Parameters for png. Default: `{'res': 300, 'width': 2000, 'height': 2000}`
		mapping: Probe to gene mapping file. If not provided, assume genes are used as rownames. This could be:
			- A column name or a number (index without samples, starting from 1) in `i.efile` to specify which column is to use a gene names
			- A file path with probe-to-gene mapping, could also include other annotations
				- A suffix(`file:1` or `file:Gene`) is available to specify the gene column. An integer indicates no header for the file, while a columname indicates `header = TRUE`
				- Without suffix, rownames will not be replaced and header is TRUE anyway.
				- Columns are ignored if this is provided.
	"""
	return Box(
		desc   = 'Detect DEGs from RNA-seq data.',
		lang   = params.Rscript.value,
		input  = "efile:file, gfile:file",
		output = """outfile:file:{{i.efile | stem | stem}}-{{i.gfile | stem
			}}.DEGs/{{i.efile | stem | stem}}-{{i.gfile | stem
			}}.degs.xls, outdir:dir:{{i.efile | stem | stem}}-{{i.gfile | stem}}.DEGs""",
		args = Box(
			tool    = 'deseq2',
			inopts  = Box(cnames = True, rnames = True, dup = 'drop'),
			mapping = "",
			cutoff  = 0.05,
			ggs = Box(
				mdsplot = Box(),
				volplot = Box(),
				maplot  = Box(),
				heatmap = Box(),
				qqplot  = Box(labs = {'x': 'Expected', 'y': 'Observed -log10(PValue)'})
			),
			params  = Box(
				volplot = Box(logfccut = 2),
				maplot = Box(),
				heatmap = Box(ngenes = None, show_row_names = False)),
			devpars = Box(res = 300, width = 2000, height = 2000)
		)
	)

@procfactory
def _pCoexp():
	"""
	@name:
		pCoexp
	@description:
		Get co-expression of gene pairs in the expression matrix.
	"""
	pCoexp             = Proc(desc = "Get co-expression of gene pairs in the expression matrix.")
	pCoexp.input       = "infile:file"
	pCoexp.output      = "outfile:file:{{i.infile | fn}}.coexp, outpval:file:{{i.infile | fn}}.pval"
	pCoexp.args.method = 'pearson'
	pCoexp.args.pval   = False
	pCoexp.lang        = params.Rscript.value
	pCoexp.script      = "file:scripts/rnaseq/pCoexp.r"
	return pCoexp

@procfactory
def _pExprSimulate():
	"""
	@name:
		pExprSimulate
	"""
	pExprSimulate               = Proc(desc = "Simulate expression values")
	pExprSimulate.input         = 'seed:var'
	pExprSimulate.output        = 'outfile:file:exprsim.{{i.seed if isinstance(i.seed, int) else "noseed"}}.txt'
	pExprSimulate.args.nsamples = 100
	pExprSimulate.args.ngenes   = 1000
	pExprSimulate.args.slabel   = 'Sample'
	pExprSimulate.args.glabel   = 'Gene'
	pExprSimulate.args.params   = Box()
	pExprSimulate.lang          = params.Rscript.value
	pExprSimulate.script        = "file:scripts/rnaseq/pExprSimulate.r"
	return pExprSimulate




from os import path
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

infile    = {{i.infile | quote}}
classfile = {{i.classfile | quote}}
outfile   = {{o.outfile | quote}}
classes   = {{args.classes | repr}}

if classfile and path.isfile(classfile):
	with open(classfile) as f:
		classes = f.read().strip().splitlines()
elif classfile:
	classes = classfile.split(',')
elif not isinstance(classes, list):
	classes = classes.split(',')

# oncotator MAFs
INTRON = "Intron"
FIVE_PRIME_UTR = "5'UTR"
THREE_PRIME_UTR = "3'UTR"
IGR = "IGR"
FIVE_PRIME_PRIME_FLANK = "5'Flank"
THREE_PRIME_PRIME_FLANK = "3'Flank"
MISSENSE = "Missense_Mutation"
NONSENSE = "Nonsense_Mutation"
NONSTOP = "Nonstop_Mutation"
SILENT = "Silent"
SPLICE_SITE = "Splice_Site"
IN_FRAME_DEL = "In_Frame_Del"
IN_FRAME_INS = "In_Frame_Ins"
FRAME_SHIFT_INS = "Frame_Shift_Ins"
FRAME_SHIFT_DEL = "Frame_Shift_Del"
START_CODON_SNP = "Start_Codon_SNP"
START_CODON_INS = "Start_Codon_Ins"
START_CODON_DEL = "Start_Codon_Del"
STOP_CODON_INS = "Stop_Codon_Ins"
STOP_CODON_DEL = "Stop_Codon_Del"
# Note: A STOP_CODON_SNP is a nonstop mutation (or Silent)
DE_NOVO_START_IN_FRAME = "De_novo_Start_InFrame"
DE_NOVO_START_OUT_FRAME = "De_novo_Start_OutOfFrame"
RNA = "RNA"
LINCRNA = "lincRNA"
# vcf2maf
TRANSLATION_START_SITE = 'Translation_Start_Site'
SPLICE_REGION = "Splice_Region"
TARGETED_REGION = "Targeted_Region"

standard_classes = [INTRON, FIVE_PRIME_UTR, THREE_PRIME_UTR, IGR, FIVE_PRIME_PRIME_FLANK, THREE_PRIME_PRIME_FLANK, MISSENSE, NONSENSE, NONSTOP, SILENT, SPLICE_SITE, IN_FRAME_DEL, IN_FRAME_INS, FRAME_SHIFT_INS, FRAME_SHIFT_DEL, START_CODON_SNP, START_CODON_INS, START_CODON_DEL, STOP_CODON_INS, STOP_CODON_DEL, DE_NOVO_START_IN_FRAME, DE_NOVO_START_OUT_FRAME, RNA, LINCRNA, TRANSLATION_START_SITE, SPLICE_REGION, TARGETED_REGION]

if any(klass not in standard_classes for klass in classes):
	raise TypeError('Not a standard variant class. Expect one of %s' % standard_classes)

reader = TsvReader(infile)
writer = TsvWriter(outfile)
writer.cnames = reader.cnames
writer.writeHead()

for r in reader:
	if r.Variant_Classification in classes:
		writer.write(r)

reader.close()
writer.close()



import sys
from cruzdb import Genome

{% if args.header %}
skip = {{args.skip}} + 1
{% else %}
skip = {{args.skip}}
{% endif %}

# get the snps
delimit = {{args.delimit | quote}}
comment = {{args.comment | quote}}
col     = {{args.col}}
with open({{i.snpfile | quote}}) as f:
	snps = sorted(set([line.split(delimit)[col] for i, line in enumerate(f.read().splitlines()) if i >= skip and line.strip() and not line.startswith(comment)]))

genome = {{args.genome | quote}}

g     = Genome (db=genome)
dbsnp = g.snp{{args.dbsnpver}}

fout  = open ("{{o.outfile}}", "w")
for snp in snps:
	s = dbsnp.filter_by(name=snp).first()
	if not s:
		sys.stderr.write('pyppl.log.warning: Cannot find coordinates for SNP: %s\n' % snp)
	else:
		# chr start end  name score strand otherinfo
		chrom   = s.chrom
		start   = s.chromStart
		end     = s.chromEnd
		name    = snp
		strand  = s.strand
		ref     = s.refUCSC
		func    = list(s.func) # intron
		alleles = [a for a in s.alleles.split(',') if a and a!=ref] or ['N']

		#         chr pos pos ref all snp comm
		fout.write ("%s %s %s %s %s %s %s\n" % (
			s.chrom,
			s.chromEnd,
			s.chromEnd,
			ref,
			','.join(alleles),
			s.name,
			"/".join(func)
		))
fout.close()


from pyppl import Box
from bioprocs.utils.tsvio import TsvReader, TsvWriter, TsvRecord

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
covfile = {{o.covfile | quote}}
cols    = {{args.cols}}
covs    = {{args.covs}}
mat     = {{args.mat | repr}}

reader = TsvReader(infile, ftype = 'head', skip = 1)
next(reader) # skip CDE_ID line

writer    = TsvWriter(outfile)
covwriter = TsvWriter(covfile)
writer.meta.add('patient', 'time', 'status')
writer.writeHead()
covwriter.meta.add('patient', *covs)
covwriter.writeHead()

realcols = {}
for key, val in cols.items():
	for v in val:
		if v in reader.meta:
			realcols[key] = v
			break

# get the sample names
samples = []
if mat:
	with open(mat) as f:
		for line in f:
			if line.startswith('#'): continue
			samples = line.strip().split('\t')
			break

def getSampleName(patient):
	ret = []
	for sample in samples:
		if sample.startswith(patient):
			ret.append(sample)
	return ret or [patient]

for r in reader:
	patient  = r[realcols.get('patient', 'patient')]
	patients = getSampleName(patient)
	status   = r[realcols.get('status', 'status')]

	if status == 'Alive':
		status = 0
	elif status == 'Dead':
		status = 1
	else:
		continue
	if status:
		time = r[realcols.get('time_death', 'time_death')]
	else:
		time = r[realcols.get('time_lastfollow', 'time_lastfollow')]
	if time == '[Not Available]' or time == '[Completed]' or time == '0':
		continue
	
	for pat in patients:
		rout         = TsvRecord()
		rout.patient = pat
		rout.time    = time
		rout.status  = status
		writer.write(rout)

		rcov         = TsvRecord()
		rcov.patient = pat
		for cov in covs:
			rcov[cov] = r[realcols.get(cov, cov)]
			if cov == 'age':
				if rcov[cov].startswith('-'):
					rcov[cov] = '{:.2f}'.format(abs(int(rcov[cov]))/365.25)
		covwriter.write(rcov)


from os import path
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter
from bioprocs.utils import logger, shell2 as shell

infile   = {{i.infile | quote}}
metafile = {{i.metafile | quote}}
outdir   = {{o.outdir | quote}}
plink    = {{args.plink | quote}}
keeptxt  = {{args.keeptxt | repr}}
chrmaps  = {{args.chrmaps | str}}
prefix   = path.join(outdir, {{i.infile | fn2 | quote}})
tpedfile  = prefix + ".tped"
tfamfile  = prefix + ".tfam"

shell.load_config(plink = plink)

# column names could be:
# FID, IID, PID, MID, Sex, Pheno
if metafile:
	logger.info('Reading metafile ...')
	metadata = dict(TsvReader(
		metafile,
		cnames = True,
		row    = lambda r: tuple((r.IID, r))
	).dump())
else:
	metadata = None

logger.info('Reading genotype matrix ...')
# snp1 gt1s1 gt1s2 ...
inreader = TsvReader(infile, cnames = True)
samples  = inreader.meta[1:]

logger.info('Writing tfam file ...')
tfamWriter = TsvWriter(tfamfile)
tfamWriter.meta = ['FID', 'IID', 'PID', 'MID', 'Sex', 'Pheno']
#tfamWriter.writeHead(callback = lambda meta: '#' + '\t'.join(meta))
uniqueIDs = {}
if not metadata:
	for s in samples:
		tfamWriter.write([s, s, '0', '0', 'other', '-9'])
else:
	for s in samples:
		fid     = metadata[s].FID if s in metadata and 'FID' in metadata[s] else s
		idcheck = fid + ' ' + s
		if idcheck in uniqueIDs:
			raise ValueError('Duplicated ID {!r}'.format(idcheck))
		uniqueIDs[idcheck] = 1
		tfamWriter.write([
			fid,
			s,
			(metadata[s].PID or '0') if s in metadata and 'PID' in metadata[s] else '0',
			(metadata[s].MID or '0') if s in metadata and 'MID' in metadata[s] else '0',
			(metadata[s].Sex or 'other') if s in metadata and 'Sex' in metadata[s] else 'other',
			(metadata[s].Pheno or '-9') if s in metadata and 'Pheno' in metadata[s] else '-9'
		])
tfamWriter.close()

def getCompondGT(gt, ref, alt):
	compGTs = {
		"0": ref + ' ' + ref,
		"1": ref + ' ' + alt,
		"2": alt + ' ' + alt
	}
	return compGTs.get(gt, '0 0')

logger.info('Writing tped file ...')
tpedWriter = TsvWriter(tpedfile)
for r in inreader:
	(chrom, pos, _, ref, alt) = r[0].split('_')
	if chrom.startswith('chr'):
		chrom = chrom[3:]
	chrom = chrmaps.get(chrom, chrom)
	tpedWriter.write([chrom, r[0], 0, pos] + [getCompondGT(gt, ref, alt) for gt in r.values()[:]])
tpedWriter.close()

logger.info("Converting using plink ...")
shell.fg.plink({'make-bed': True}, tfile = prefix, out = prefix)

if not keeptxt:
	shell.rm_rf(tpedfile)
	shell.rm_rf(tfamfile)



from os import makedirs, path
from shutil import rmtree
from pyppl import Box
from bioprocs.utils import cmdargs, runcmd, mem2

tmpdir    = path.join ("{{args.tmpdir}}", "{{proc.id}}.{{i.infile | fn}}.{{job.index}}")
if not path.exists (tmpdir): makedirs (tmpdir)

mem    = mem2({{ args.mem | quote }}, 'java')
ref    = {{args.ref | quote}}
params = {{args.params}}
try:
	params['TMP_DIR'] = tmpdir
	params['I']       = {{i.infile | quote}}
	params['O']       = {{o.outfile | quote}}
	params['R']       = ref
	runcmd ('{{args.picard}} ReorderSam %s -Djava.io.tmpdir="%s" %s' % (mem, tmpdir, cmdargs(params, dash = '', equal = '=')))
except Exception as ex:
	stderr.write ("Job failed: %s" % str(ex))
	raise
finally:
	rmtree (tmpdir)



"""
Read or write MEME motif file
"""
import re
import math
from collections import OrderedDict

class MemeRecord(object):

	def __init__(self, 
		name, 
		matrix,
		altname = '',
		mtype   = 'letter-probability',
		alength = None,
		w       = None,
		nsites  = 20,
		E       = 0,
		URL     = None
	):
		self.name    = name
		self.matrix  = matrix
		self.altname = altname
		self.mtype   = mtype
		self.alength = alength or len(matrix[0])
		self.w       = w or len(matrix)
		self.nsites  = nsites
		self.E       = E
		self.URL     = URL

	def __str__(self):
		return """
MOTIF {name}{altname}
{mtype} matrix: alength= {alength} w= {w} nsites= {nsites} E= {E}
{matrix}
{URL}
""".format(
		name    = self.name,
		altname = " " + self.altname if self.altname else "",
		mtype   = self.mtype,
		alength = self.alength,
		w       = self.w,
		nsites  = self.nsites,
		E       = self.E,
		matrix  = "\n".join(" ".join(str(r) for r in row) for row in self.matrix),
		URL     = "URL {}".format(self.URL) if self.URL else ""
	)

	def pwm2logodds(self):
		assert self.mtype == 'letter-probability'
		matrix = [
			tuple(math.exp(p)/(1.0 + math.exp(p)) for p in row)
			for row in self.matrix
		]
		return MemeRecord(
			name    = self.name,
			matrix  = matrix,
			altname = self.altname,
			mtype   = 'log-odds',
			alength = self.alength,
			w       = self.w,
			nsites  = self.nsites,
			E       = self.E,
			URL     = self.URL
		)

	def pwm2prob(self):
		assert self.mtype == 'log-odds'
		matrix = [
			tuple(math.log(p/(1.0-p)) for p in row)
			for row in self.matrix
		]
		return MemeRecord(
			name    = self.name,
			matrix  = matrix,
			altname = self.altname,
			mtype   = 'letter-probability',
			alength = self.alength,
			w       = self.w,
			nsites  = self.nsites,
			E       = self.E,
			URL     = self.URL
		)

class MemeReader(object):

	def __init__(self, memefile):
		self.meta     = {}
		alphabet_flag = False
		bgfreqs_flag  = False
		self.file = open(memefile)
		self.tell = 0
		while True:
			self.tell = self.file.tell()
			line = self.file.readline()
			if not line:
				raise ValueError('Not a valid MEME motif file.')
			if line.startswith('MEME version'):
				self.meta['Version'] = line[12:].strip()
			elif line.startswith('ALPHABET='):
				self.meta['Alphabet'] = line[9:].strip()
			elif line.startswith('ALPHABET'):
				self.meta['Alphabet'] = line[8:].strip()
				alphabet_flag = True
			elif line.startswith('END ALPHABET'):
				alphabet_flag = False
			elif alphabet_flag:
				self.meta['Alphabet'] += '\n' + line.strip()
			elif line.startswith('strands:'):
				self.meta['Strands'] = line[8:].strip()
			elif line.startswith('Background letter frequencies'):
				bgfreqs_flag = True
				source = line[30:].strip()
				if source.startswith('(from '):
					source = source[6:-2]
				else:
					source = ''
				self.meta['bgfreqs'] = {'from': source, 'freqs': OrderedDict()}
			elif bgfreqs_flag:
				bgfreqs_flag = False
				parts = line.strip().split()
				self.meta['bgfreqs']['freqs'] = OrderedDict(tuple([parts[2*i], float(parts[2*i+1])] for i  in range(int(len(parts)/2))))
			elif line.startswith('MOTIF'):
				self.file.seek(self.tell)
				break

	def next(self):
		name    = None
		altname = ''
		url     = None
		mtype   = ''
		matrix  = []
		attrs   = {}
		while True:
			tell = self.file.tell()
			line = self.file.readline()
			if not line:
				raise StopIteration()
			if line.startswith('MOTIF'):
				if name:
					self.file.seek(tell)
					break
				parts = line[5:].strip().split()
				name  = parts.pop(0)
				if parts:
					altname = parts[0]
			elif line.startswith('URL'):
				url = line[3:].strip()
			elif 'matrix:' in line:
				matrix = [] # in case there are multiple matrices
				mtype, attrs = line.strip().split('matrix:')
				mtype = mtype.strip()
				attrs = re.split(r'(?:\s*=\s*|\s+)', attrs.strip())
				attrs = {attrs[2*i]:attrs[2*i+1] for i in range(int(len(attrs)/2))}
			else:
				line = line.strip()
				if not line:
					continue
				matrix.append(tuple(float(v) for v in line.split()))
		
		return MemeRecord(
			name, 
			matrix,
			altname = altname,
			mtype   = mtype,
			URL     = url,
			**attrs
		)
	
	def __next__(self):
		return self.next()

	def rewind(self):
		self.file.seek(self.tell)

	def __iter__(self):
		return self

	def __del__(self):
		self.close()

	def close(self):
		if self.file:
			self.file.close()

class MemeWriter(object):

	def __init__(self, outfile, meta = None):
		self.meta = meta or {}
		self.file = open(outfile, 'w')

	def writeMeta(self):
		self.file.write("MEME version {}\n\n".format(self.meta.get('Version', 4)))
		alphabet = self.meta.get('Alphabet', 'ACGT')
		if '\n' in alphabet:
			self.file.write("ALPHABET {}\nEND ALPHABET\n\n".format(alphabet))
		else:
			self.file.write("ALPHABET= {}\n\n".format(alphabet))
		strands = self.meta.get('Strands', '+ -')
		self.file.write("strands: {}\n\n".format(strands))
		bgfreqs = self.meta.get("bgfreqs", {})
		if "from" in bgfreqs:
			self.file.write("Background letter frequencies (from {}):\n".format(bgfreqs['from']))
		if "freqs" in bgfreqs:
			self.file.write(" ".join('{} {}'.format(k, v) for k, v in bgfreqs['freqs'].items()) + "\n\n")
	
	def write(self, mrec):
		self.file.write(str(mrec))

	def __del__(self):
		self.close()

	def close(self):
		if self.file:
			self.file.close()



{{dsnparse}}
{{schemaparse}}

from sys import stderr
dsn = dsnparse({{i.dsn | quote}})
if not hasattr(dsn, 'scheme'):
	raise Exception('Cannot determine the scheme from DSN string.')

# maybe needed for other scheme
name, fields = schemaparse({{i.datafile | quote}}, type = 'data', scheme = dsn.scheme, delimit = {{args.delimit | quote}})
name = {{i.datafile | tablename | quote}}

if dsn.scheme in ['sqlite', 'sqlite3']:
	assert hasattr(dsn, 'file') and dsn.file != ':memory:'
	import sqlite3
	conn = sqlite3.connect(dsn.file)
	conn.text_factory = str

	# drop table
	{% if args.drop %}
	dropsql = 'DROP TABLE IF EXISTS "%s"' % name
	conn.execute(dropsql)
	conn.commit()
	{% endif %}

	# create table
	createsql  = 'CREATE TABLE IF NOT EXISTS "%s" (\n' % name
	for i, field in enumerate(fields):
		fname, ftype, fstat = field
		if i == len(fields) - 1:
			createsql += '  "%s" %s %s\n' % (fname, ftype, fstat)
		else:
			createsql += '  "%s" %s %s,\n' % (fname, ftype, fstat)
	createsql += ') %s' % dsn.after

	stderr.write('Executing SQL:\n')
	stderr.write(createsql + '\n')
	conn.execute(createsql)
	conn.commit()

	insertsql = 'INSERT INTO "%s" ("%s") VALUES (%s);' % (name, '", "'.join([
		field[0] for field in fields
	]), ', '.join(['?' for _ in fields]))
	data = []
	with open({{i.datafile | quote}}, 'r') as f:
		f.readline() # skip header
		for line in f:
			line = line.strip('\n')
			if not line: continue
			line = line.split({{args.delimit | quote}})
			data.append(tuple(line))
	if data:
		stderr.write('Executing SQL:\n')
		stderr.write(insertsql + '\n')
		conn.executemany(insertsql, data)
	conn.commit()
	conn.close()
else:
	raise NotImplementedError('Only support sqlite3 currently.')


import json, glob, os
from shutil import copyfile
from bioprocs.utils.parallel import Parallel
from threading import Lock

# load varialbes
indir   = {{i.indir | quote}}
mdfile  = {{i.mdfile | quote}}
outdir  = {{o.outdir | quote}}
nthread = {{args.nthread | repr}}
method  = {{args.method | quote}}

sam_meta = None
sample_ids = {}
with open (mdfile) as f:
	sam_meta = json.load (f)

exts = dict()
for sam in sam_meta:
	parts = sam['file_name'].split('.')
	ext = '.' + parts[-1]
	if ext == '.gz':
		ext = '.' + parts[-2] + ext
	exts      [sam['file_name']] = ext
	sample_ids[sam['file_name']] = sam['associated_entities'][0]['entity_submitter_id'][:15]

samfiles = []
for ext in set(exts.values()):
	samfiles += glob.glob (os.path.join(os.path.abspath(indir), "*" + ext))
	# or direct dir from TCGA download
	samfiles += glob.glob (os.path.join(os.path.abspath(indir), "*", "*" + ext))

lock = Lock()
def single(samfile):
	bn = os.path.basename (samfile)
	if not bn in sample_ids: return
	newfile = os.path.join (outdir, sample_ids[bn] + exts[bn])
	with lock:
		if os.path.exists (newfile):
			os.remove(newfile)
		if 'link' in method:
			os.symlink (samfile, newfile)
		elif method == 'copy':
			copyfile(samfile, newfile)

p = Parallel(nthread = nthread, backend = 'threading', raiseExc = True)
p.run(single, [(samfile,) for samfile in samfiles])


from os import path
from pyppl import Box
from bioprocs.utils import shell2 as shell

cnvkit   = {{args.cnvkit | quote}}
infiles  = {{i.infiles | repr}}
ref      = {{args.ref | quote}}
outfile  = {{o.outfile | quote}}
params   = {{args.params | repr}}
nthread  = {{args.nthread | repr}}

shell.load_config(cnvkit = dict(
	_exe = cnvkit,
	_env = dict(
		OPENBLAS_NUM_THREADS = str(nthread),
		OMP_NUM_THREADS      = str(nthread),
		NUMEXPR_NUM_THREADS  = str(nthread),
		MKL_NUM_THREADS      = str(nthread)
	),
	_cwd = path.dirname(outfile)
))

params.o = outfile
params.f = ref
shell.fg.cnvkit.reference(*infiles, **params)



#!/usr/bin/env python
"""Do PWM scan"""
from pyppl import PyPPL, Box
from bioprocs import params

params._prefix = '-'

params.tflist.show     = True
params.tfmotifs.show   = True
params.consvdir.show   = True

params.input           = "tf"
params.input.desc      = "The input content. tf or motif."

params.infile.required = True
params.infile.desc     = "The input file. Whether TF or motif list."

params.target          = 'gene'
params.target.desc     = "What's the target. gene or region"

params.tfile.required  = True
params.tfile.desc      = 'The target file. If target is region, using BED format.'

params.pval            = 1e-4
params.pval.desc       = "The p-value cutoff for PWM scan."

params.consvp          = 5e-2
params.consvp.desc     = "The conservation p-value cutoff. 0 to disable conservation report."

params.outdir.required = True
params.outdir.desc     = "The output directory."

params.tfnorm          = True
params.tfnorm.desc     = 'Whether normalize TF names.'

params.nthread         = 1
params.nthread.desc    = 'Number of threads used for scanning.'

def main():
	opts = params._parse(dict_wrapper = Box)

	from bioprocs.gene import pGeneNameNorm
	from bioprocs.sets.tfbs import aTfbsTfPC, aTfbsTfRC, aTfbsTfP, aTfbsTfR, \
		aTfbsPC, aTfbsRC, aTfbsP, aTfbsR

	# check params
	if not opts.input in ['tf', 'motif']:
		raise ValueError('Unexpected "input" type, should be one of ["tf", "motif"]')

	if not opts.target in ['gene', 'region']:
		raise ValueError('Unexpected "target" type, should be one of ["gene", "region"]')

	if opts.tfnorm and opts.input == 'tf':
		pGeneNameNorm.input = [opts.infile]
		PyPPL().start(pGeneNameNorm).run()
		opts.infile = pGeneNameNorm.channel.get()

	if opts.input == 'tf' and opts.target == 'gene' and opts.consvp > 0:
		aggr = aTfbsTfPC
	elif opts.input == 'tf' and opts.target == 'gene' and opts.consvp == 0:
		aggr = aTfbsTfP
	elif opts.input == 'tf' and opts.target == 'region' and opts.consvp > 0:
		aggr = aTfbsTfRC
	elif opts.input == 'tf' and opts.target == 'region' and opts.consvp == 0:
		aggr = aTfbsTfR
	if opts.input == 'motif' and opts.target == 'gene' and opts.consvp > 0:
		aggr = aTfbsPC
	elif opts.input == 'motif' and opts.target == 'gene' and opts.consvp == 0:
		aggr = aTfbsP
	elif opts.input == 'motif' and opts.target == 'region' and opts.consvp > 0:
		aggr = aTfbsRC
	elif opts.input == 'motif' and opts.target == 'region' and opts.consvp == 0:
		aggr = aTfbsR

	if opts.input == 'tf':
		aggr.input = [[opts.infile], [opts.tfile], [opts.tflist]]
	else:
		aggr.input = [[opts.infile], [opts.tfile]]

	if opts.consvp > 0:
		aggr.args.cpval = opts.consvp
		aggr.pConsvPerm.args.consvdir = opts.consvdir

	aggr.args.pval               = opts.pval
	aggr.args.tfmotifs           = opts.tfmotifs
	aggr.acache                  = True
	aggr.exdir                   = opts.outdir
	aggr.pMotifScan.args.nthread = opts.nthread

	PyPPL().start(aggr).run()

if __name__ == "__main__":
	main()








from bioprocs.utils.shell2 import bgzip

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
gz      = {{args.gz | bool}}
if gz:
	outfile = outfile[:-3]

with open(infile) as fin, open(outfile, 'w') as fout:
	for line in fin:
		line = line.rstrip('\n')
		if line.startswith('#'):
			fout.write(line + '\n')
			continue

		parts = line.split('\t')
		for i, part in enumerate(parts):
			formats = part.split(':')
			if formats[0] != '0/0' and formats[0] != '0|0':
				continue
			if any(fmt != '.' for fmt in formats[1:]):
				continue
			formats[0] = './.'
			parts[i] = ':'.join(formats)
		fout.write('\t'.join(parts) + '\n')

if gz:
	bgzip(outfile)



"""A set of resources to download via API or URL links"""
from pyppl import Proc, Box
from . import params
#from .utils import txt, download, runcmd
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pTxt():
	"""
	@name:
		pTxt
	@description:
		Download CSV format files.
	@input:
		`in`: The name of the resource
	@output:
		`outfile:file`: The output file
	@args:
		`cols`:      Select the columns to keep. Default: '' (all cols)
		`rowfilter`: Filter rows. For example, to filter out rows not start with 'Chr':
			- `"lambda x: not x[0].startswith('Chr')"`
			- Note that rowfilter applied before cols filter.
		`urls`:      Available resources and their urls.
		`gz`:        Whether to gzip the output file.
	@requires:
		[`curl`](https://en.wikipedia.org/wiki/CURL)
	"""
	pTxt                      = Proc(desc = 'Download CSV format files.')
	pTxt.input                = "in"
	pTxt.output               = "outfile:file:{{i.in}}.txt{{args.gz | lambda x: '.gz' if x else ''}}"
	pTxt.args.gz              = False
	pTxt.args.delimit         = "\t"
	pTxt.args.skip            = 0
	pTxt.args.cols            = ''
	pTxt.args.header          = True
	pTxt.args.rowfilter       = ''
	pTxt.args.transform       = ''
	pTxt.args.username        = ''
	pTxt.args.password        = ''
	pTxt.args.curl            = params.curl.value
	# pTxt.tplenvs.txtFilter    = txt.filter.py
	# pTxt.tplenvs.txtTransform = txt.transform.py
	# pTxt.tplenvs.downloadCurl = download.curl.py
	# pTxt.tplenvs.runcmd       = runcmd.py
	pTxt.args.urls            = Box({
		'drugbank-target-all': 'https://www.drugbank.ca/releases/5-0-7/downloads/target-all-uniprot-links',
		'drugbank-target-approved': 'https://www.drugbank.ca/releases/5-0-7/downloads/target-approved-uniprot-links',
		'ccle-sample-info': 'https://data.broadinstitute.org/ccle_legacy_data/cell_line_annotations/CCLE_sample_info_file_2012-10-18.txt',
		'ccle-rseq-rpkm': 'https://data.broadinstitute.org/ccle/CCLE_RNAseq_081117.rpkm.gct',
		'ccle-rseq-reads': 'https://data.broadinstitute.org/ccle/CCLE_RNAseq_081117.reads.gct',
		'KEGG_2016_gmt': 'http://amp.pharm.mssm.edu/Enrichr/geneSetLibrary?mode=text&libraryName=KEGG_2016',
		'GO_Molecular_Function_2017_gmt': 'http://amp.pharm.mssm.edu/Enrichr/geneSetLibrary?mode=text&libraryName=GO_Molecular_Function_2017',
		'GO_Cellular_Component_2017_gmt': 'http://amp.pharm.mssm.edu/Enrichr/geneSetLibrary?mode=text&libraryName=GO_Cellular_Component_2017',
		'GO_Biological_Process_2017_gmt': 'http://amp.pharm.mssm.edu/Enrichr/geneSetLibrary?mode=text&libraryName=GO_Biological_Process_2017',
		'TargetScan_microRNA': 'http://amp.pharm.mssm.edu/Enrichr/geneSetLibrary?mode=text&libraryName=TargetScan_microRNA',
		'TRANSFAC_and_JASPAR_PWMs': 'http://amp.pharm.mssm.edu/Enrichr/geneSetLibrary?mode=text&libraryName=TRANSFAC_and_JASPAR_PWMs',
	})
	pTxt.lang   = params.python.value
	pTxt.script = "file:scripts/resource/pTxt.py"
	return pTxt

@procfactory
def _pGtf():
	"""
	@name:
		pGtf
	"""
	pGtf                      = Proc(desc = 'Download GTF files.')
	pGtf.input                = "in"
	pGtf.output               = "outfile:file:{{in}}.gtf{{args.gz | lambda x: '.gz' if x else ''}}"
	pGtf.args.gz              = False
	pGtf.args.curl            = 'curl'
	pGtf.args.username        = ''
	pGtf.args.password        = ''
	pGtf.args.genepredtogtf   = 'genePredToGtf'
	# pGtf.tplenvs.runcmd       = runcmd.py
	# pGtf.tplenvs.downloadCurl = download.curl.py
	pGtf.args.urls            = {
		'hg19-refgene': 'http://hgdownload.cse.ucsc.edu/goldenPath/hg19/database/refGene.txt.gz',
		'hg19-knowngene': 'http://hgdownload.cse.ucsc.edu/goldenPath/hg19/database/knownGene.txt.gz',
		'hg38-refgene': 'http://hgdownload.cse.ucsc.edu/goldenPath/hg38database/refGene.txt.gz',
		'hg38-knowngene': 'http://hgdownload.cse.ucsc.edu/goldenPath/hg38/database/knownGene.txt.gz',
	}
	pGtf.lang   = params.python.value
	pGtf.script = """
	import os, shutil
	from subprocess import check_output
	{{downloadCurl}}
	{{runcmd}}
	url = {{args.urls | json}}["{{in}}"]
	tmpdir = "{{job.outdir}}/tmp"
	if not os.path.exists(tmpdir):
		os.makedirs(tmpdir)

	downfile = os.path.join(tmpdir, 'downloaded')
	downloadCurl(url, downfile, {{args.username | quote}}, {{args.password | quote}}, {{args.curl | quote}})
	outfile = "{{outfile}}"[:-3] if {{args.gz | bool}} else "{{outfile}}"
	output  = check_output(['file', downfile])
	if 'gzip' in output:
		ugfile = downfile + '.ungz'
		with open(ugfile, 'w') as f:
			f.write(check_output(['gunzip', downfile, '-c']))
		downfile = ugfile
	elif 'Zip' in output:
		zipdir = os.path.join(tmpdir, '_unzipped')
		import zipfile, glob
		zipref = zipfile.ZipFile(downfile, 'r')
		zipref.extractall(zipdir)
		zipref.close()
		downfile = glob.glob(os.path.join(zipdir, '*'))[0]
	cutfile = downfile + '.cutf2'
	with open(downfile) as fin, open(cutfile, 'w') as fout:
		for line in fin:
			if not line.strip(): continue
			fout.write("\\t".join(line.split("\\t")[1:]))
	runcmd ('{{args.genepredtogtf}} file "%s" "%s" -source="{{in}}"' % (cutfile, outfile))
	if {{args.gz | bool}}:
		runcmd ('gz "%s"' % outfile)
	shutil.rmtree(tmpdir)
	"""
	return pGtf




"""Putative transcription factor binding sites analysis"""
from pyppl import Proc, Box
from . import params
from .utils import fs2name
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pMotifScan():
	"""
	@name:
		pMotifScan
	@description:
		Scan motif along the given sequences.
	@input:
		`tffile:file`: The infile containing TF name and motif name.
			- If only one column is give, will be used as both TF and motif name
			- If there are 2+ columns, 1st column will be motif name, 2nd column will be TF name
		`sfile:file`: The sequence file
	@output:
		`outdir:file`: The output dir
	@args:
		`tools`   : The tool used to scan the motif. Default: 'meme'
		`meme`    : The path of MEME's fimo. Default: 'fimo'
		`motifs`  : The motif database in MEME format.
		`pval`    : The pvalue cutoff. Default: 1e-4
		`cleanmname`: Whether to clean motif name. Default: True
		`ucsclink`: The ucsc link template. Default: `https://genome.ucsc.edu/cgi-bin/hgTracks?db=hg19&position={}`
		`nthread` : Number of threads used to scan, only available when you have multiple mids. Default: 1
		`params`  : Other parameters for `fimo`
	@requires:
		[`fimo` from MEME Suite](http://meme-suite.org/tools/fimo)
	"""
	pMotifScan                        = Proc(desc = 'Scan motif along the given sequences.')
	pMotifScan.input                  = "tffile:file, sfile:file"
	pMotifScan.output                 = [
		"outfile:file:{{i.sfile | fn}}-{{i.tffile | fn}}.fimo/{{i.sfile | fn}}-{{i.tffile | fn}}.bed",
		"outdir:dir:{{i.sfile | fn}}-{{i.tffile | fn}}.fimo"
	]
	pMotifScan.args.tool     = 'meme'
	pMotifScan.args.meme     = params.fimo.value
	pMotifScan.args.params   = Box()
	pMotifScan.args.tfmotifs = params.tfmotifs.value
	pMotifScan.args.pval     = 1e-4
	pMotifScan.args.ucsclink = 'https://genome.ucsc.edu/cgi-bin/hgTracks?db=hg19&position={}'
	pMotifScan.args.nthread  = 1
	pMotifScan.lang          = params.python.value
	pMotifScan.script        = "file:scripts/tfbs/pMotifScan.py"
	return pMotifScan

@procfactory
def _pMotifSimilarity():
	"""
	@name:
		pMotifSimilarity
	@description:
		Compare the similarity between motifs.
	@input:
		`mfile1:file`: The motif database in meme format
		`mfile2:file`: Another motif database in meme format.
			- If not provided, `mfile1` will be used.
	@output:
		`outfile:file`: The output file from `tomtom`
		`outdir:dir`  : The output directory from `tomtom`
	@args:
		`qval`   : The qvalue cutoff. Default: `0.5`
		`tomtom` : The path to `tomtom`
		`nthread`: Number of threads to use. Default: `1`
		`params` : parameters for `tomtom`:
			- `xalph`       : `True`
			- `no-ssc`      : `True`
			- `dist`        : `pearson`,
			- `min-overlap` : `1`,
			- `motif-pseudo`: `.1`,
			- `verbosity`   : `4`
	"""
	pMotifSimilarity        = Proc(desc = 'Compare the similarity between motifs.')
	pMotifSimilarity.input  = 'mfile1:file, mfile2:file'
	pMotifSimilarity.output = [
		'outfile:file:{{odfn(i.mfile1, i.mfile2, fn2) | lambda x, path=path: path.join(x, "tomtom.tsv")}}',
		'outdir:dir:{{odfn(i.mfile1, i.mfile2, fn2)}}'
	]
	pMotifSimilarity.args.qval   = 0.5
	pMotifSimilarity.args.tomtom = params.tomtom.value
	pMotifSimilarity.args.params = Box({
		'xalph'       : True,
		'no-ssc'      : True,
		'dist'        : 'pearson',
		'min-overlap' : 1,
		'motif-pseudo': .1,
		'verbosity'   : 4
	})
	pMotifSimilarity.args.nthread = 1
	pMotifSimilarity.envs.path    = __import__('os').path
	pMotifSimilarity.envs.odfn    = lambda f1, f2, fn2: (fn2(f1), fn2(f1)+"-"+fn2(f2))[int(bool(f2))] + ".tomtom"
	pMotifSimilarity.lang         = params.python.value
	pMotifSimilarity.script       = "file:scripts/tfbs/pMotifSimilarity.py"
	return pMotifSimilarity

@procfactory
def _pMotifMerge():
	"""
	@name:
		pMotifMerge
	"""
	pMotifMerge              = Proc(desc = 'Merge motif files in MEME format')
	pMotifMerge.input        = 'infiles:files'
	pMotifMerge.output       = 'outfile:file:{{i.infiles | fs2name}}.meme.txt'
	pMotifMerge.envs.fs2name = fs2name
	pMotifMerge.lang         = params.python.value
	pMotifMerge.script       = "file:scripts/tfbs/pMotifMerge.py"
	return pMotifMerge

@procfactory
def _pMotifFilter():
	"""
	@name:
		pMotifFilter
	@description:
		Filter motifs from a meme file.
	@input:
		`infile:file`: The motif database in MEME format
	@output:
		`outfile:file`: the output file.
	@args:
		`filter`: A string of lambda function to filter the motifs.
			- Possible attributes for a motif are:
			- `E, URL, alength, altname, matrix, mtype, name, nsites, w`
			- Motifs will be filtered out when this function returns `True`.
	"""
	pMotifFilter             = Proc(desc = 'Filter motifs from a meme file.')
	pMotifFilter.input       = 'infile:file'
	pMotifFilter.output      = 'outfile:file:{{i.infile | bn}}'
	pMotifFilter.args.filter = None
	pMotifFilter.lang        = params.python.value
	pMotifFilter.script      = "file:scripts/tfbs/pMotifFilter.py"
	return pMotifFilter

@procfactory
def _pAtSnp():
	"""
	@name:
		pAtSnp
	@description:
		Scan motifs on Snps to detect binding affinity changes.
	@input:
		`tffile:file`:  The tf-motif file with 1st column the motif and 2nd the tf
		`snpfile:file`: The snp file.
			- Could be a bed file with first 6 columns the bed6 format and 7th the reference allele, and 8th the alternative alleles.
			- Alleles including reference allele should be seperated by `,`
			- It also could be a snp file required by `atSNP` package.
			- See: https://rdrr.io/github/kinsigne/atSNP_modified/man/LoadSNPData.html
	@output:
		`outfile:file`: The output file
		`outdir:dir`  : The output directory containing the output file and plots.
	@args:
		`tfmotifs`: The motif database. Defaut: `params.tfmotifs`
		`genome`  : The reference genome to get the sequences. Default: `params.genome`
		`fdr`     : Do fdr or not. Or could be the p-value adjustment method. Default: `True` (using `BH` method)
		`pval`    : The pvalue cutoff for output and plot.
		`plot`    : Do plot or not. Default: `True`
		`nthread` : How many threads to use. Default: `1`
		`depvars` : The device parameters for plotting. Default: `Box(res = 300, width = 2000, height = 2000)`
	@requires:
		`r-atSNP`
	"""
	pAtSnp               = Proc(desc = 'Scan motifs on Snps to detect binding affinity changes.')
	pAtSnp.input         = 'tffile:file, snpfile:file'
	pAtSnp.output        = [
		'outfile:file:{{i.tffile | fn2}}-{{i.snpfile | fn2}}.atsnp/{{i.tffile | fn2}}-{{i.snpfile | fn2}}.atsnp.txt',
		'outdir:dir:{{i.tffile | fn2}}-{{i.snpfile | fn2}}.atsnp'
	]
	pAtSnp.args.tfmotifs = params.tfmotifs.value
	pAtSnp.args.genome   = params.genome.value
	pAtSnp.args.fdr      = True
	pAtSnp.args.pval     = 0.05
	pAtSnp.args.plot     = True
	pAtSnp.args.nthread  = 1
	pAtSnp.args.devpars  = Box(res = 300, width = 2000, height = 2000)
	pAtSnp.lang          = params.Rscript.value
	pAtSnp.script        = "file:scripts/tfbs/pAtSnp.r"
	return pAtSnp





# outfile should be:
"""
cluster	gene	is.driver	P.vaf	R.vaf	P.ccf	R.ccf	P.ref.count	P.var.count	P.depth	R.ref.count	R.var.count	R.depth	P	R
1	SMC3	TRUE	50.28	49.13	100.56	98.26	3027	2611	5639	2035	1990	4024	50.28	49.13
1	PTPRT	TRUE	45.24	45.31	90.48	90.62	2159	1799	3958	1715	1579	3294	45.24	45.31
1	-	FALSE	43.83	41.37	87.66	82.74	2931	2364	5296	2224	1501	3725	43.83	41.37
"""

# pyclone loci file
"""
mutation_id	sample_id	cluster_id	cellular_prevalence	cellular_prevalence_std	variant_allele_frequency
chr10:17875816	NA12878	2	0.828865827591	0.0355432431918	0.904883227176
chr10:17875816	NA12156	2	0.741264908504	0.11050033704	0.560311284047
chr10:17875816	NA19240	2	0.837320950405	0.0572230410102	0.89578313253
"""

# ref/alt counts from pyclone's yaml file:
"""
mutations:
- id: chr10:91143374
  ref_counts: 5506
  states:
  - g_n: AA
    g_r: AA
    g_v: AB
    prior_weight: 1
  - g_n: AA
    g_r: AA
    g_v: BB
    prior_weight: 1
  var_counts: 33
"""
from pathlib import Path
import yaml
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter, TsvRecord
from bioprocs.utils import shell2 as shell
indir    = Path({{i.indir | quote}})
outfile  = Path({{o.outfile | quote}})
refgene  = {{args.refgene | quote}}
bedtools = {{args.bedtools | quote}}
drivers  = {{args.drivers | str}}
shell.load_config(bedtools = bedtools)

# get samples
samples = [path.stem for path in (indir/'yaml').glob('*.yaml')]

# get basic mutation information from loci file
locifile = indir / 'tables' / 'loci.tsv'
locireader = TsvReader(locifile)
records = {}
for loci in locireader:
	if loci.mutation_id not in records:
		records[loci.mutation_id] = TsvRecord()
	records[loci.mutation_id][loci.sample_id + '.vaf'] = loci.variant_allele_frequency
	records[loci.mutation_id][loci.sample_id] = loci.variant_allele_frequency
	# is it ok to use cellular_prevalence as ccf?
	records[loci.mutation_id][loci.sample_id + '.ccf'] = loci.cellular_prevalence
	records[loci.mutation_id]['cluster'] = int(loci.cluster_id) + 1

# get the counts from yaml file
for ymlfile in (indir/'yaml').glob('*.yaml'):
	with ymlfile.open() as ymlstrm:
		mutations = yaml.safe_load(ymlstrm)
		for mutation in mutations['mutations']:
			if mutation['id'] not in records:
				records[mutation['id']] = TsvRecord()
			records[mutation['id']][ymlfile.stem + '.ref.count'] = mutation['ref_counts']
			records[mutation['id']][ymlfile.stem + '.var.count'] = mutation['var_counts']
			records[mutation['id']][ymlfile.stem + '.depth'] = int(mutation['ref_counts']) + int(mutation['var_counts'])

# annotate mutations with genes
# construct mutation bed file
bedwriter = TsvWriter(outfile.parent / 'mutations.bed')
for mutation in records:
	chrom, start = mutation.split(':')
	end = int(start) + 1
	bedwriter.write([chrom, start, end, mutation, 0, '+'])
bedwriter.close()

mutations_with_genes = outfile.parent / 'mutations_with_genes.txt'
shell.bedtools.intersect(a = bedwriter.filename, b = refgene, wo = True, _out = str(mutations_with_genes))

bedreader = TsvReader(mutations_with_genes)
for bedr in bedreader:
	if bedr[3] not in records:
		continue
	# gene_id "IFIT1B"; transcript_id "IFIT1B"; gene_name "IFIT1B";
	gene = bedr[14].split('; ')[0][8:].strip('"')
	if 'gene' in records[bedr[3]]:
		records[bedr[3]].gene += ',' + gene
	else:
		records[bedr[3]].gene = gene
	if gene in drivers:
		records[bedr[3]]["is.driver"] = 'TRUE'

writer = TsvWriter(outfile)
writer.cnames = ['cluster', 'gene', 'is.driver']
for suffix in ('.vaf', '.ccf', '.ref.count', '.var.count', '.depth', ''):
	for sample in samples:
		writer.cnames.append(sample + suffix)
writer.writeHead()
# fill the defaults
for mutation, record in records.items():
	if 'cluster' not in record:
		#record['cluster'] = 'NA'
		# skip mutations with cluster
		continue
	if 'gene' not in record:
		record['gene'] = '-'
	if 'is.driver' not in record:
		record['is.driver'] = 'FALSE'
	for sample in samples:
		for suffix in ('.vaf', '.ccf', '.ref.count', '.var.count', '.depth', ''):
			if sample + suffix not in record:
				record[sample + suffix] = 'NA'
	writer.write(record)
writer.close()



"""A set of processes handling Excel files (>=2010)"""

from pyppl import Proc
from . import params
from .utils import fs2name
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pTsvs2Xlsx():
	"""
	@name:
		pTsvs2Xlsx
	@description:
		Save tsv files to xlsx sheets.
	@input:
		`infiles:files`: The input tsv files
	@output:
		`outfile:file`: The output xlsx file
	@args:
		`fn2sheet`: How to convert filename(without extension) to sheet name
	@requires:
		python packages: `csv` and `openpyxl`
	"""
	pTsvs2Xlsx               = Proc(desc = 'Save tsv files to xlsx sheets.')
	pTsvs2Xlsx.input         = 'infiles:files'
	pTsvs2Xlsx.output        = 'outfile:file:{{i.infiles | fs2name}}.xlsx'
	pTsvs2Xlsx.args.fn2sheet = None
	pTsvs2Xlsx.envs.fs2name  = fs2name
	pTsvs2Xlsx.lang          = params.python.value
	pTsvs2Xlsx.script        = "file:scripts/xlsx/pTsvs2Xlsx.py"
	return pTsvs2Xlsx




"""Some misc processes"""
from pyppl import Proc, Box
from bioprocs import params, rimport
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pGEP70():
	"""
	@name:
		pGEP70
	@description:
		Calculate GEP70 scores for multiple mylenoma 70-gene-signatures and do survival plot.
		Or add one gene to it to see the survival plot.
		see: https://www.ncbi.nlm.nih.gov/pubmed/17105813
	@input:
		`exprfile:file`: The input file with expression matrix
			- Columns are samples, rows are genes
			- make sure the expression values are log2-scale normalized
		`survfile:file`: The survival data file (header is required).
			- col1: rownames
			- col2: the survival time
			- col3: the status. 0/1 for alive/dead or 1/2 for alive dead
		`gene`: An extra gene to be added to the plot.
			- If not provided, only do the GEP70 plot.
	@output:
		`outdir:file`: The output directory, containing:
			- The survival data file.
			- The GEP70 plot and results
			- The GEP70 + gene plot and results if `i.gene` is provided.
	@args:
		`gep70`: The GEP70 genes.
			- Column 1: genes
			- Column 2: how is the regulated (up or down)
		`inunit`  : The time unit in input file. Default: days
		`outunit` : The output unit for plots. Default: days
		`params`  : The params for `ggsurvplot`. Default: `Box({'risk.table': True, 'conf.int': True, 'font.legend': 13, 'pval': 'Log-rank p = {pval}'})`
			- You may do `ylim.min` to set the min ylim. Or you can set it as 'auto'. Default: 0.
		`ggs`     : Extra ggplot2 elements for main plot. `ggs.table` is for the risk table.
		`devpars` : The device parameters for png. Default: `{res:300, height:2000, width:2000}`
	"""
	pGEP70              = Proc(desc = 'Do GEP70 plot for multiple mylenoma 70-gene-signatures')
	pGEP70.input        = 'exprfile:file, survfile:file, gene'
	pGEP70.output       = 'outdir:dir:{{i.survfile | fn2}}.{{args.name}}{{i.gene}}'
	pGEP70.args.name    = 'GEP70'
	pGEP70.args.gep70   = params.gep70.value
	pGEP70.args.inunit  = 'days' # months, weeks, years
	pGEP70.args.outunit = 'days'
	pGEP70.args.params  = Box({'font.legend': 13, 'pval': 'Log-rank p = {pval}', 'risk.table': True})
	pGEP70.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pGEP70.args.ggs     = Box(table = Box())
	pGEP70.envs.rimport = rimport
	pGEP70.lang         = 'Rscript'
	pGEP70.script       = "file:scripts/misc/pGEP70.r"
	return pGEP70

@procfactory
def _pNCBI():
	"""
	@name:
		pNCBI
	@description:
		The NCBI E-Utils
	@input:
		`term`: The term or the id argument for esearch or efetch
	@output:
		`outfile:file`: The output file
	@args:
		`prog`   : The program to use, esearch (Default) or efetch
		`apikey` : The api key for E-utils
			- Without API key, we can only query 3 time in a second
			- With it, we can do 10.
		`sleep`  : Sleep sometime after job done. Default: `0.15`
			- Because of the limit of # queries/sec, we need to sleep sometime after the job is done
			- At the same time, we also have to limit # jobs to run at the same time. typically: `pNCBI.forks = 10`
		`db`     : The database to query. Default: `pubmed`. Available databases:
			- annotinfo, assembly, biocollections, bioproject, biosample, biosystems, blastdbinfo, books,
			- cdd, clinvar, clone, dbvar, gap, gapplus, gds, gencoll, gene, genome, geoprofiles, grasp, gtr,
			- homologene, ipg, medgen, mesh, ncbisearch, nlmcatalog, nuccore, nucest, nucgss, nucleotide,
			- omim, orgtrack, pcassay, pccompound, pcsubstance, pmc, popset, probe, protein, proteinclusters,
			- pubmed, pubmedhealth, seqannot, snp, sparcle, sra, structure, taxonomy, unigene
		`joiner` : The delimit to use if the field is a list
		`record` : A function to transform the record.
	@requires:
		[python-eutils](https://github.com/biocommons/eutils)
	"""
	pNCBI             = Proc(desc = 'The NCBI E-Utils')
	pNCBI.input       = 'term'
	pNCBI.output      = 'outfile:file:{{i.term | lambda x: __import__("re").sub(r"[^\\w_]", "_", x)[:255]}}.{{args.prog}}.txt'
	pNCBI.errhow      = 'retry'
	pNCBI.args.prog   = 'esearch'
	pNCBI.args.apikey = params.ncbikey.value
	pNCBI.args.sleep  = .15
	# annotinfo, assembly, biocollections, bioproject, biosample, biosystems, blastdbinfo, books,
	# cdd, clinvar, clone, dbvar, gap, gapplus, gds, gencoll, gene, genome, geoprofiles, grasp, gtr,
	# homologene, ipg, medgen, mesh, ncbisearch, nlmcatalog, nuccore, nucest, nucgss, nucleotide,
	# omim, orgtrack, pcassay, pccompound, pcsubstance, pmc, popset, probe, protein, proteinclusters,
	# pubmed, pubmedhealth, seqannot, snp, sparcle, sra, structure, taxonomy, unigene
	pNCBI.args.db     = 'pubmed'
	pNCBI.args.joiner = '|'
	pNCBI.args.record = None
	pNCBI.lang        = params.python.value
	pNCBI.script      = 'file:scripts/misc/pNCBI.py'
	return pNCBI




import re
from os import path, symlink
from sys import stdout, stderr
from time import sleep
from shutil import move
from pyppl import Box
from bioprocs.utils import runcmd, cmdargs, reference, logger
from bioprocs.utils.poll import Poll

# variables
infile   = {{i.fq | quote}}
outfile  = {{o.outfile | quote}}
outfmt   = {{args.outfmt | quote}}
outdir   = {{job.outdir | quote}}
ref      = {{args.ref | quote}}
refgene  = {{args.refgene | quote}}
ref2     = path.join(outdir, path.basename(ref))
workdir  = {{proc.workdir | quote}}
rg       = {{args.rg}}
params   = {{args.params}}
samtools = {{args.samtools | quote}}
bowtie2  = {{args.bowtie2 | quote}}
bwa      = {{args.bwa | quote}}
ngm      = {{args.ngm | quote}}
star     = {{args.star | quote}}
nthread  = {{args.nthread}}
joblen   = {{proc.size}}
jobindex = {{job.index}}
bowtie2_build  = {{args.bowtie2_build | quote}}
poll     = Poll(workdir, joblen, jobindex)

# determine whether the reference file is specified
logger.info('Checking if reference exists: %s' % ref)
reference.check(ref)

# detemine default read group
rg = {key.upper():val for key, val in rg.items()}
if not rg['ID']:
	g = re.search (r'[^a-zA-Z0-9]+(L\\d+)[^a-zA-Z0-9]+', {{o.outfile | fn | quote}})
	rg['ID'] = g.group(1) if g else "{{o.outfile | fn}}.L{{job.index}}"
if not rg['SM']:
	rg['SM'] = {{o.outfile | fn | quote}}

def sam2bam(samfile, bamfile):
	logger.info('Converting sam to bam: ')
	logger.info('- %s' % samfile)
	logger.info('- %s' % bamfile)
	cmd = '%s view -Sb "%s" > "%s"; rm -f "%s"' % (samtools, samfile, bamfile, samfile)
	runcmd(cmd)

def checkAndBuildIndex(indexes, buildcmd, buildcmd2):
	def todo(indexes, ref, buildcmd, ref2, buildcmd2):
		logger.info('Checking if reference file is indexed.')
		if not reference.checkIndex(indexes):
			logger.info('No, trying to build index at the first job.')
			reference.buildIndex(ref, buildcmd, ref2, buildcmd2)
	poll.first(todo, indexes, ref, buildcmd, ref2, buildcmd2)
	return ref2 if path.exists(ref2) else ref

############ bowtie2
{% case args.tool %}
{% when 'bowtie2' %}
# check reference index files
indexes = [
	"%s.1.bt2" % ref,
	"%s.2.bt2" % ref,
	"%s.3.bt2" % ref,
	"%s.4.bt2" % ref,
	"%s.rev.1.bt2" % ref,
	"%s.rev.2.bt2" % ref
]
buildcmdbase = "{bowtie2_build} --thread {nthread} '{ref}' '{ref}'"
buildcmd     = buildcmdbase.format(
	bowtie2_build = bowtie2_build, nthread = nthread, ref = ref
)
buildcmd2    = buildcmdbase.format(
	bowtie2_build = bowtie2_build, nthread = nthread, ref = ref2
)
ref = checkAndBuildIndex(indexes, buildcmd, buildcmd2)

# do mapping
params['threads'] = nthread
params['rg-id']   = rg['ID']
i = 0
for k,v in rg.items():
	if k == 'ID': continue
	params['rg' + ' '*i] = k + ':' + v
	i += 1
params['x'] = ref
params['U'] = infile
params['S'] = path.splitext(outfile)[0] + '.sam'

cmd = '{bowtie2} {args}'.format(bowtie2 = bowtie2, args = cmdargs(params, equal = ' '))
logger.info('Do mapping ...')
runcmd (cmd)
if outfmt == 'bam':	sam2bam(params['S'], outfile)

############ bwa
{% when 'bwa' %}
indexes = [
	"%s.bwt" % ref,
	"%s.amb" % ref,
	"%s.ann" % ref,
	"%s.pac" % ref,
	"%s.sa" % ref,
]
buildcmdbase = "{bwa} index '{ref}'"
buildcmd     = buildcmdbase.format(bwa = bwa, ref = ref)
buildcmd2    = buildcmdbase.format(bwa = bwa, ref = ref2)
ref = checkAndBuildIndex(indexes, buildcmd, buildcmd2)

# do mapping
params['t'] = nthread
params['R'] = "@RG\\tID:%s\\t%s" % (rg['ID'], "\\t".join(k + ':' +v for k,v in rg.items() if k!='ID'))
outfile2    = path.splitext(outfile)[0] + '.sam'
cmd         = "{bwa} mem {args} '{ref}' '{infile}' > '{outfile}'".format(
	bwa     = bwa,
	args    = cmdargs(params),
	ref     = ref,
	infile  = infile,
	outfile = outfile2
)
logger.info('Do mapping ...')
runcmd (cmd)
if outfmt == 'bam':	sam2bam(outfile2, outfile)

{% when 'ngm' %}
indexes = [
	"%s-enc.2.ngm" % ref,
	"%s-ht-13-2.3.ngm" % ref
]
buildcmdbase = "{ngm} -r '{ref}' -t {nthread}"
buildcmd     = buildcmdbase.format(ngm = ngm, ref = ref, nthread = nthread)
buildcmd2    = buildcmdbase.format(ngm = ngm, ref = ref2, nthread = nthread)
ref = checkAndBuildIndex(indexes, buildcmd, buildcmd2)

# do mapping
params['q'] = infile
params['r'] = ref
params['o'] = outfile
params['b'] = {{ args.outfmt | lambda x: x=='bam' }}
params.update({('rg-' + k.lower()):v for k,v in rg.items()})
params['t'] = nthread
cmd = '%s %s' % (ngm, cmdargs(params, equal = ' '))
logger.info('Do mapping ...')
runcmd (cmd)

############ star
{% when 'star' %}
indexes = [path.join(path.splitext(ref)[0] + '.star', x) for x in [
	"chrLength.txt",
	"chrNameLength.txt",
	"chrName.txt",
	"chrStart.txt",
	"exonGeTrInfo.tab",
	"exonInfo.tab",
	"geneInfo.tab",
	"Genome",
	"genomeParameters.txt",
	"SA",
	"SAindex",
	"sjdbInfo.txt",
	"sjdbList.fromGTF.out.tab",
	"sjdbList.out.tab",
	"transcriptInfo.tab"
]]
buildcmdbase = "mkdir -p '{stardir}'; {star} --runMode genomeGenerate --genomeDir '{stardir}' --genomeFastaFiles " \
               + "'{ref}' --sjdbOverhang 100 --sjdbGTFfile '{refgene}' --runThreadN {nthread}"
buildcmd = buildcmdbase.format(
	stardir = path.splitext(ref)[0] + '.star',
	star    = star,
	ref     = ref,
	refgene = refgene,
	nthread = nthread
)
buildcmd2 = buildcmdbase.format(
	stardir = path.splitext(ref2)[0] + '.star',
	star    = star,
	ref     = ref2,
	refgene = refgene,
	nthread = nthread
)
ref = checkAndBuildIndex(indexes, buildcmd, buildcmd2)
refDir = path.splitext(ref)[0] + '.star'
rfcmd  = "zcat" if infile.endswith('.gz') else 'bzcat' if infile.endswith('.bz2') else 'cat'

params['genomeDir']         = refDir
params['readFilesIn']       = infile
params['readFilesCommand']  = rfcmd
params['readNameSeparator'] = '.'
params['outFileNamePrefix'] = outdir + '/'
params['outSAMtype']        = [outfmt.upper(), 'Unsorted']
cmd                         = '%s %s' % (star, cmdargs(params, equal = ' '))
logger.info('Do mapping ...')
runcmd (cmd)

outfile2 = "{outdir}/Aligned.out.{outfmt}".format(outdir = outdir, outfmt = outfmt)
if path.exists(outfile2): move(outfile2, outfile)
{% endcase %}
